from copy import deepcopy
from dataclasses import dataclass
from typing import Mapping, Any, Optional, List, Union, Dict, FrozenSet, Tuple, TYPE_CHECKING

from dbt.adapters.base.impl import AdapterConfig, ConstraintSupport
from dbt.adapters.base.meta import available
from dbt.adapters.capability import CapabilityDict, CapabilitySupport, Support, Capability
from dbt.adapters.catalogs import CatalogRelation, CatalogIntegration, CatalogIntegrationConfig
from dbt.adapters.contracts.relation import RelationConfig
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.events.types import ColTypeChange
from dbt.adapters.cache import _make_ref_key_dict
from dbt.adapters.sql.impl import (
    LIST_SCHEMAS_MACRO_NAME,
    LIST_RELATIONS_MACRO_NAME,
)
from dbt_common.contracts.constraints import ConstraintType
from dbt_common.contracts.metadata import (
    TableMetadata,
    StatsDict,
    StatsItem,
    CatalogTable,
    ColumnMetadata,
)
from dbt_common.events.functions import fire_event
from dbt_common.exceptions import CompilationError, DbtDatabaseError, DbtRuntimeError
from dbt_common.utils import filter_null_values

from dbt.adapters.snowflake import constants, parse_model
from dbt.adapters.snowflake.catalogs import (
    BuiltInCatalogIntegration,
    InfoSchemaCatalogIntegration,
    IcebergRestCatalogIntegration,
)
from dbt.adapters.snowflake.relation_configs import SnowflakeRelationType

from dbt.adapters.snowflake import SnowflakeColumn
from dbt.adapters.snowflake import SnowflakeConnectionManager
from dbt.adapters.snowflake import SnowflakeRelation

if TYPE_CHECKING:
    import agate

SHOW_OBJECT_METADATA_MACRO_NAME = "snowflake__show_object_metadata"


@dataclass
class SnowflakeConfig(AdapterConfig):
    transient: Optional[bool] = None
    partition_by: Optional[Union[str, List[str]]] = None
    cluster_by: Optional[Union[str, List[str]]] = None
    automatic_clustering: Optional[bool] = None
    secure: Optional[bool] = None
    copy_grants: Optional[bool] = None
    snowflake_warehouse: Optional[str] = None
    query_tag: Optional[str] = None
    tmp_relation_type: Optional[str] = None
    merge_update_columns: Optional[str] = None
    target_lag: Optional[str] = None
    row_access_policy: Optional[str] = None
    table_tag: Optional[str] = None

    # extended formats
    table_format: Optional[str] = None
    external_volume: Optional[str] = None
    base_location_root: Optional[str] = None
    base_location_subpath: Optional[str] = None


class SnowflakeAdapter(SQLAdapter):
    Relation = SnowflakeRelation
    Column = SnowflakeColumn
    ConnectionManager = SnowflakeConnectionManager

    AdapterSpecificConfigs = SnowflakeConfig

    CATALOG_INTEGRATIONS = [
        BuiltInCatalogIntegration,
        InfoSchemaCatalogIntegration,
        IcebergRestCatalogIntegration,
    ]
    CONSTRAINT_SUPPORT = {
        ConstraintType.check: ConstraintSupport.NOT_SUPPORTED,
        ConstraintType.not_null: ConstraintSupport.ENFORCED,
        ConstraintType.unique: ConstraintSupport.NOT_ENFORCED,
        ConstraintType.primary_key: ConstraintSupport.NOT_ENFORCED,
        ConstraintType.foreign_key: ConstraintSupport.NOT_ENFORCED,
    }

    _capabilities: CapabilityDict = CapabilityDict(
        {
            Capability.SchemaMetadataByRelations: CapabilitySupport(support=Support.Full),
            Capability.TableLastModifiedMetadata: CapabilitySupport(support=Support.Full),
            Capability.TableLastModifiedMetadataBatch: CapabilitySupport(support=Support.Full),
            Capability.GetCatalogForSingleRelation: CapabilitySupport(support=Support.Full),
            Capability.MicrobatchConcurrency: CapabilitySupport(support=Support.Full),
        }
    )

    def __init__(self, config, mp_context) -> None:
        super().__init__(config, mp_context)
        self.add_catalog_integration(constants.DEFAULT_INFO_SCHEMA_CATALOG)
        self.add_catalog_integration(constants.DEFAULT_BUILT_IN_CATALOG)

    def add_catalog_integration(
        self, catalog_integration: CatalogIntegrationConfig
    ) -> CatalogIntegration:
        # don't mutate the object that dbt-core passes in
        catalog_integration = deepcopy(catalog_integration)
        catalog_integration.name = catalog_integration.name.upper()
        return super().add_catalog_integration(catalog_integration)

    def get_catalog_integration(self, name: str) -> CatalogIntegration:
        # Snowflake uppercases everything in their metadata tables
        return super().get_catalog_integration(name.upper())

    @classmethod
    def date_function(cls):
        return "CURRENT_TIMESTAMP()"

    @classmethod
    def _catalog_filter_table(
        cls, table: "agate.Table", used_schemas: FrozenSet[Tuple[str, str]]
    ) -> "agate.Table":
        # On snowflake, users can set QUOTED_IDENTIFIERS_IGNORE_CASE, so force
        # the column names to their lowercased forms.
        lowered = table.rename(column_names=[c.lower() for c in table.column_names])
        return super()._catalog_filter_table(lowered, used_schemas)

    def _make_match_kwargs(self, database, schema, identifier):
        # if any path part is already quoted then consider same casing but without quotes
        quoting = self.config.quoting
        if self._is_quoted(identifier):
            identifier = self._strip_quotes(identifier)
        elif identifier is not None and quoting["identifier"] is False:
            identifier = identifier.upper()

        if self._is_quoted(schema):
            schema = self._strip_quotes(schema)
        elif schema is not None and quoting["schema"] is False:
            schema = schema.upper()

        if self._is_quoted(database):
            database = self._strip_quotes(database)
        elif database is not None and quoting["database"] is False:
            database = database.upper()

        return filter_null_values(
            {"identifier": identifier, "schema": schema, "database": database}
        )

    def _is_quoted(self, identifier: str) -> bool:
        return (
            identifier is not None
            and identifier.startswith(self.Relation.quote_character)
            and identifier.endswith(self.Relation.quote_character)
        )

    def _strip_quotes(self, identifier: str) -> str:
        return identifier.strip(self.Relation.quote_character)

    def _get_warehouse(self) -> str:
        _, table = self.execute("select current_warehouse() as warehouse", fetch=True)
        if len(table) == 0 or len(table[0]) == 0:
            # can this happen?
            raise DbtRuntimeError("Could not get current warehouse: no results")
        return str(table[0][0])

    def _use_warehouse(self, warehouse: str):
        """Use the given warehouse. Quotes are never applied."""
        self.execute("use warehouse {}".format(warehouse))

    def pre_model_hook(self, config: Mapping[str, Any]) -> Optional[str]:
        default_warehouse = self.config.credentials.warehouse
        warehouse = config.get("snowflake_warehouse", default_warehouse)
        if warehouse == default_warehouse or warehouse is None:
            return None
        previous = self._get_warehouse()
        self._use_warehouse(warehouse)
        return previous

    def post_model_hook(self, config: Mapping[str, Any], context: Optional[str]) -> None:
        if context is not None:
            self._use_warehouse(context)

    def list_schemas(self, database: str) -> List[str]:
        try:
            results = self.execute_macro(LIST_SCHEMAS_MACRO_NAME, kwargs={"database": database})
        except DbtDatabaseError as exc:
            msg = f"Database error while listing schemas in database " f'"{database}"\n{exc}'
            raise DbtRuntimeError(msg)
        # this uses 'show terse schemas in database', and the column name we
        # want is 'name'

        return [row["name"] for row in results]

    def get_columns_in_relation(self, relation):
        try:
            return super().get_columns_in_relation(relation)
        except DbtDatabaseError as exc:
            if "does not exist or not authorized" in str(exc):
                return []
            else:
                raise

    def _show_object_metadata(self, relation: SnowflakeRelation) -> Optional[dict]:
        try:
            kwargs = {"relation": relation}
            results = self.execute_macro(SHOW_OBJECT_METADATA_MACRO_NAME, kwargs=kwargs)

            if len(results) == 0:
                return None

            return results
        except DbtDatabaseError:
            return None

    def get_catalog_for_single_relation(
        self, relation: SnowflakeRelation
    ) -> Optional[CatalogTable]:
        object_metadata = self._show_object_metadata(relation.as_case_sensitive())

        if not object_metadata:
            return None

        row = object_metadata[0]

        is_dynamic = row.get("is_dynamic") in ("Y", "YES")
        kind = row.get("kind")

        if is_dynamic and kind == str(SnowflakeRelationType.Table).upper():
            table_type = str(SnowflakeRelationType.DynamicTable).upper()
        else:
            table_type = kind

        # https://docs.snowflake.com/en/sql-reference/sql/show-views#output
        # Note: we don't support materialized views in dbt-snowflake
        is_view = kind == str(SnowflakeRelationType.View).upper()

        table_metadata = TableMetadata(
            type=table_type,
            schema=row.get("schema_name"),
            name=row.get("name"),
            database=row.get("database_name"),
            comment=row.get("comment"),
            owner=row.get("owner"),
        )

        stats_dict: StatsDict = {
            "has_stats": StatsItem(
                id="has_stats",
                label="Has Stats?",
                value=True,
                include=False,
                description="Indicates whether there are statistics for this table",
            ),
            "row_count": StatsItem(
                id="row_count",
                label="Row Count",
                value=row.get("rows"),
                include=(not is_view),
                description="Number of rows in the table as reported by Snowflake",
            ),
            "bytes": StatsItem(
                id="bytes",
                label="Approximate Size",
                value=row.get("bytes"),
                include=(not is_view),
                description="Size of the table as reported by Snowflake",
            ),
        }

        catalog_columns = {
            c.column: ColumnMetadata(type=c.dtype, index=i + 1, name=c.column)
            for i, c in enumerate(self.get_columns_in_relation(relation))
        }

        return CatalogTable(
            metadata=table_metadata,
            columns=catalog_columns,
            stats=stats_dict,
        )

    def list_relations_without_caching(
        self, schema_relation: SnowflakeRelation
    ) -> List[SnowflakeRelation]:
        kwargs = {"schema_relation": schema_relation}

        try:
            schema_objects = self.execute_macro(LIST_RELATIONS_MACRO_NAME, kwargs=kwargs)
        except DbtDatabaseError as exc:
            # if the schema doesn't exist, we just want to return.
            # Alternatively, we could query the list of schemas before we start
            # and skip listing the missing ones, which sounds expensive.
            # "002043 (02000)" is error code for "object does not exist or is not found"
            # The error message text may vary across languages, but the error code is expected to be more stable
            if "002043 (02000)" in str(exc):
                return []
            raise

        columns = ["database_name", "schema_name", "name", "kind", "is_dynamic", "is_iceberg"]
        schema_objects = schema_objects.rename(
            column_names=[col.lower() for col in schema_objects.column_names]
        )
        return [self._parse_list_relations_result(obj) for obj in schema_objects.select(columns)]

    def _parse_list_relations_result(self, result: "agate.Row") -> SnowflakeRelation:
        database, schema, identifier, relation_type, is_dynamic, is_iceberg = result

        try:
            relation_type = self.Relation.get_relation_type(relation_type.lower())
        except ValueError:
            relation_type = self.Relation.External

        if relation_type == self.Relation.Table and is_dynamic == "Y":
            relation_type = self.Relation.DynamicTable

        table_format = (
            constants.ICEBERG_TABLE_FORMAT
            if is_iceberg in ("Y", "YES")
            else constants.INFO_SCHEMA_TABLE_FORMAT
        )

        quote_policy = {"database": True, "schema": True, "identifier": True}

        return self.Relation.create(
            database=database,
            schema=schema,
            identifier=identifier,
            type=relation_type,
            table_format=table_format,
            quote_policy=quote_policy,
        )

    def quote_seed_column(self, column: str, quote_config: Optional[bool]) -> str:
        quote_columns: bool = False
        if isinstance(quote_config, bool):
            quote_columns = quote_config
        elif quote_config is None:
            pass
        else:
            msg = (
                f'The seed configuration value of "quote_columns" has an '
                f"invalid type {type(quote_config)}"
            )
            raise CompilationError(msg)

        if quote_columns:
            return self.quote(column)
        else:
            return column

    @available
    def standardize_grants_dict(self, grants_table: "agate.Table") -> dict:
        grants_dict: Dict[str, Any] = {}

        for row in grants_table:
            grantee = row["grantee_name"]
            granted_to = row["granted_to"]
            privilege = row["privilege"]
            if privilege != "OWNERSHIP" and granted_to not in ["SHARE", "DATABASE_ROLE"]:
                if privilege in grants_dict.keys():
                    grants_dict[privilege].append(grantee)
                else:
                    grants_dict.update({privilege: [grantee]})
        return grants_dict

    def timestamp_add_sql(self, add_to: str, number: int = 1, interval: str = "hour") -> str:
        return f"DATEADD({interval}, {number}, {add_to})"

    def submit_python_job(self, parsed_model: dict, compiled_code: str):
        schema = parsed_model["schema"]
        database = parsed_model["database"]
        identifier = parsed_model["alias"]
        python_version = parsed_model["config"].get(
            "python_version", constants.DEFAULT_PYTHON_VERSION_FOR_PYTHON_MODELS
        )

        packages = parsed_model["config"].get("packages", [])
        imports = parsed_model["config"].get("imports", [])
        external_access_integrations = parsed_model["config"].get(
            "external_access_integrations", []
        )
        secrets = parsed_model["config"].get("secrets", {})
        # adding default packages we need to make python model work
        default_packages = ["snowflake-snowpark-python"]
        package_names = [package.split("==")[0] for package in packages]
        for default_package in default_packages:
            if default_package not in package_names:
                packages.append(default_package)
        packages = "', '".join(packages)
        imports = "', '".join(imports)
        external_access_integrations = ", ".join(external_access_integrations)
        secrets = ", ".join(f"'{key}' = {value}" for key, value in secrets.items())

        # we can't pass empty imports, external_access_integrations or secrets clause to snowflake
        if imports:
            imports = f"IMPORTS = ('{imports}')"
        if external_access_integrations:
            # Black is trying to make this a tuple.
            # fmt: off
            external_access_integrations = f"EXTERNAL_ACCESS_INTEGRATIONS = ({external_access_integrations})"
        if secrets:
            secrets = f"SECRETS = ({secrets})"

        if self.config.args.SEND_ANONYMOUS_USAGE_STATS:
            snowpark_telemetry_string = "dbtLabs_dbtPython"
            snowpark_telemetry_snippet = f"""
import sys
sys._xoptions['snowflake_partner_attribution'].append("{snowpark_telemetry_string}")"""
        else:
            snowpark_telemetry_snippet = ""

        common_procedure_code = f"""
RETURNS STRING
LANGUAGE PYTHON
RUNTIME_VERSION = '{python_version}'
PACKAGES = ('{packages}')
{external_access_integrations}
{secrets}
{imports}
HANDLER = 'main'
EXECUTE AS CALLER
AS
$$
{snowpark_telemetry_snippet}

{compiled_code}
$$"""

        use_anonymous_sproc = parsed_model["config"].get("use_anonymous_sproc", True)
        if use_anonymous_sproc:
            proc_name = f"{identifier}__dbt_sp"
            python_stored_procedure = f"""
WITH {proc_name} AS PROCEDURE ()
{common_procedure_code}
CALL {proc_name}();
            """
        else:
            proc_name = f"{database}.{schema}.{identifier}__dbt_sp"
            python_stored_procedure = f"""
CREATE OR REPLACE PROCEDURE {proc_name} ()
{common_procedure_code};
CALL {proc_name}();

            """
        response, _ = self.execute(python_stored_procedure, auto_begin=False, fetch=False)
        if not use_anonymous_sproc:
            self.execute(
                f"drop procedure if exists {proc_name}()",
                auto_begin=False,
                fetch=False,
            )
        return response

    def valid_incremental_strategies(self):
        return ["append", "merge", "delete+insert", "microbatch", "insert_overwrite"]

    def debug_query(self):
        """Override for DebugTask method"""
        self.execute("select 1 as id")

    @classmethod
    def _get_adapter_specific_run_info(cls, config: RelationConfig) -> Dict[str, Any]:
        # `config` is not a RelationConfig!
        run_info = {
            "adapter_type": constants.ADAPTER_TYPE,
            "table_format": None,
        }

        if config and hasattr(config, "_extra"):

            catalog = config._extra.get("catalog")

            if _table_format := config._extra.get("table_format"):  # type:ignore
                run_info["table_format"] = _table_format
            elif not catalog:
                # no table_format and no catalog definitely means info schema table
                run_info["table_format"] = constants.INFO_SCHEMA_TABLE_FORMAT
            elif catalog == constants.DEFAULT_INFO_SCHEMA_CATALOG.name:  # type:ignore
                # if the user happens to set the catalog to the info schema catalog, catch that
                run_info["table_format"] = constants.INFO_SCHEMA_TABLE_FORMAT
            elif catalog == constants.DEFAULT_ICEBERG_REST_CATALOG.name:  # type:ignore
                # if the user happens to set the catalog to the iceberg rest catalog, catch that
                run_info["table_format"] = constants.ICEBERG_TABLE_FORMAT
            else:  # catalog is set, and it's not the info schema catalog
                # it's unlikely that users will set a catalog that's not Iceberg
                run_info["table_format"] = constants.ICEBERG_TABLE_FORMAT

        return run_info

    @available
    def build_catalog_relation(self, model: RelationConfig) -> Optional[CatalogRelation]:
        """
        Builds a relation for a given configuration.

        This method uses the provided configuration to determine the appropriate catalog
        integration and config parser for building the relation. It defaults to the built-in Iceberg
        catalog if none is provided in the configuration for backward compatibility.

        Args:
            model (RelationConfig): `config.model` (not `model`) from the jinja context

        Returns:
            Any: The constructed relation object generated through the catalog integration and parser
        """
        if catalog := parse_model.catalog_name(model):
            catalog_integration = self.get_catalog_integration(catalog)
            return catalog_integration.build_relation(model)
        return None

    @available
    def describe_dynamic_table(self, relation: SnowflakeRelation) -> Dict[str, Any]:
        """
        Get all relevant metadata about a dynamic table to return as a dict to Agate Table row

        Args:
            relation (SnowflakeRelation): the relation to describe
        """
        quoting = relation.quote_policy
        schema = f'"{relation.schema}"' if quoting.schema else relation.schema
        database = f'"{relation.database}"' if quoting.database else relation.database
        show_sql = (
            f"show dynamic tables like '{relation.identifier}' in schema {database}.{schema}"
        )
        res, dt_table = self.execute(show_sql, fetch=True)
        if res.code != "SUCCESS":
            raise DbtRuntimeError(f"Could not get dynamic query metadata: {show_sql} failed")
        # normalize column names to lower case, this still preserves column order
        dt_table = dt_table.rename(column_names=[name.lower() for name in dt_table.column_names])
        return {
            "dynamic_table": dt_table.select(
                [
                    "name",
                    "schema_name",
                    "database_name",
                    "text",
                    "target_lag",
                    "warehouse",
                    "refresh_mode",
                ]
            )
        }

    def expand_column_types(self, goal, current):
        reference_columns = {c.name: c for c in self.get_columns_in_relation(goal)}

        target_columns = {c.name: c for c in self.get_columns_in_relation(current)}

        for column_name, reference_column in reference_columns.items():
            target_column = target_columns.get(column_name)

            if target_column is not None and target_column.can_expand_to(reference_column):
                col_string_size = reference_column.string_size()
                new_type = self.Column.string_type(col_string_size)
                if collation := target_column.collation:
                    new_type += f"collate '{collation}'"
                fire_event(
                    ColTypeChange(
                        orig_type=target_column.data_type,
                        new_type=new_type,
                        table=_make_ref_key_dict(current),
                    )
                )

                self.alter_column_type(current, column_name, new_type)



from dbt_common.record import record_function

from dbt.adapters.record import RecordReplayCursor
from dbt.adapters.snowflake.record.cursor.sfqid import CursorGetSfqidRecord
from dbt.adapters.snowflake.record.cursor.sqlstate import CursorGetSqlStateRecord
from dbt.adapters.snowflake.record.cursor.stats import CursorGetStatsRecord


class SnowflakeRecordReplayCursor(RecordReplayCursor):
    """A custom extension of RecordReplayCursor that adds the sqlstate,
    sfqid, and stats properties which are specific to snowflake-connector."""

    @property
    @record_function(CursorGetSqlStateRecord, method=True, id_field_name="connection_name")
    def sqlstate(self):
        return self.native_cursor.sqlstate

    @property
    @record_function(CursorGetSfqidRecord, method=True, id_field_name="connection_name")
    def sfqid(self):
        return self.native_cursor.sfqid

    @property
    @record_function(CursorGetStatsRecord, method=True, id_field_name="connection_name")
    def stats(self):
        return self.native_cursor.stats



import os

import pytest

from dbt.tests.util import run_dbt, run_dbt_and_capture

"""
Testing rationale:
- snowflake SHOW TERSE OBJECTS command returns at max 10K objects in a single call
- when dbt attempts to write into a schema with more than 10K objects, compilation will fail
  unless we paginate the result
- we default pagination to 10 pages, but users want to configure this
  - we instead use that here to force failures by making it smaller
"""


TABLE = """
{{ config(materialized='table') }}
select 1 as id
"""


VIEW = """
{{ config(materialized='view') }}
select id from {{ ref('my_model_base') }}
"""


DYNAMIC_TABLE = (
    """
{{ config(
    materialized='dynamic_table',
    target_lag='1 hour',
    snowflake_warehouse='"""
    + os.getenv("SNOWFLAKE_TEST_WAREHOUSE", "")
    + """',
) }}

select id from {{ ref('my_model_base') }}
"""
)


class BaseConfig:
    VIEWS = 90
    DYNAMIC_TABLES = 10

    @pytest.fixture(scope="class")
    def models(self):
        my_models = {"my_model_base.sql": TABLE}
        for view in range(self.VIEWS):
            my_models[f"my_model_{view}.sql"] = VIEW
        for dynamic_table in range(self.DYNAMIC_TABLES):
            my_models[f"my_dynamic_table_{dynamic_table}.sql"] = DYNAMIC_TABLE
        return my_models

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["run"])

    def test_list_relations(self, project):
        kwargs = {"schema_relation": project.test_schema}
        with project.adapter.connection_named("__test"):
            relations = project.adapter.execute_macro(
                "snowflake__list_relations_without_caching", kwargs=kwargs
            )
        assert len(relations) == self.VIEWS + self.DYNAMIC_TABLES + 1

    def test_on_run(self, project):
        _, logs = run_dbt_and_capture(["run"])
        assert "list_relations_per_page" not in logs
        assert "list_relations_page_limit" not in logs


class TestListRelationsWithoutCachingSmall(BaseConfig):
    pass


class TestListRelationsWithoutCachingLarge(BaseConfig):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "flags": {
                "list_relations_per_page": 10,
                "list_relations_page_limit": 20,
            }
        }


class TestListRelationsWithoutCachingTooLarge(BaseConfig):
    """
    We raise a warning, not an error, if the number of relations exceeds the page limit times the iteration limit.
    In this case, we return the maximum number of relations (e.g. 5 pages * 10 relations per page) and continue.
    """

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "flags": {
                "list_relations_per_page": 10,
                "list_relations_page_limit": 5,
            }
        }

    def test_list_relations(self, project):
        kwargs = {"schema_relation": project.test_schema}
        with project.adapter.connection_named("__test"):
            relations = project.adapter.execute_macro(
                "snowflake__list_relations_without_caching", kwargs=kwargs
            )
            assert (
                len(relations) == 10 * 5
            )  # the maximum from the settings, not the number of relations (101)

    def test_on_run(self, project):
        # the warning should only show in this scenario
        _, logs = run_dbt_and_capture(["run"])
        assert "list_relations_per_page" in logs
        assert "list_relations_page_limit" in logs



import pytest
import uuid
from dbt.tests.util import run_dbt, write_file
from dbt.tests.adapter.python_model.test_python_model import (
    BasePythonEmptyTests,
    BasePythonModelTests,
    BasePythonIncrementalTests,
    BasePythonSampleTests,
)

models__simple_python_model = """
import pandas

def model(dbt, session):
    dbt.config(
        materialized='table',
    )
    data = [[1,2]] * 10
    return pandas.DataFrame(data, columns=['test', 'test2'])
"""

models__simple_python_model_v2 = """
import pandas

def model(dbt, session):
    dbt.config(
        materialized='table',
    )
    data = [[1,2]] * 10
    return pandas.DataFrame(data, columns=['test1', 'test3'])
"""

models__custom_target_model = """
import pandas

def model(dbt, session):
    dbt.config(
        materialized="table",
        schema="MY_CUSTOM_SCHEMA",
        alias="_TEST_PYTHON_MODEL",
    )

    df = pandas.DataFrame({
        'City': ['Buenos Aires', 'Brasilia', 'Santiago', 'Bogota', 'Caracas'],
        'Country': ['Argentina', 'Brazil', 'Chile', 'Colombia', 'Venezuela'],
        'Latitude': [-34.58, -15.78, -33.45, 4.60, 10.48],
        'Longitude': [-58.66, -47.91, -70.66, -74.08, -66.86]
    })

    return df
"""


class TestPythonModelSnowflake(BasePythonModelTests):
    pass


class TestIncrementalSnowflake(BasePythonIncrementalTests):
    pass


class TestIncrementalSnowflakeQuoting(BasePythonModelTests):
    # ensure that 'dbt.ref()', 'dbt.this()', and py_write_table() all respect quoting
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"quoting": {"identifier": True}}


class TestChangingSchemaSnowflake:
    @pytest.fixture(scope="class")
    def models(self):
        return {"simple_python_model.py": models__simple_python_model}

    def test_changing_schema(self, project):
        run_dbt(["run"])
        write_file(
            models__simple_python_model_v2,
            project.project_root + "/models",
            "simple_python_model.py",
        )
        run_dbt(["run"])


USE_IMPORT_MODEL = """
import sys
from snowflake.snowpark.types import StructType, FloatType, StringType, StructField

def model( dbt, session):

    dbt.config(
        materialized='table',
        imports = ['@dbt_integration_test/iris.csv'],
        use_anonymous_sproc = False
    )
    schema_for_data_file = StructType([
        StructField("length1", FloatType()),
        StructField("width1", FloatType()),
        StructField("length2", FloatType()),
        StructField("width2", FloatType()),
        StructField("variety", StringType()),
    ])
    df = session.read.schema(schema_for_data_file).option("field_delimiter", ",").schema(schema_for_data_file).csv("@dbt_integration_test/iris.csv")
    return df
"""


class TestImportSnowflake:
    @pytest.fixture(scope="class")
    def models(self):
        return {"simple_python_model.py": USE_IMPORT_MODEL}

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"iris.csv": "1,2,3,4,setosa"}

    def test_import(self, project):
        project.run_sql("create or replace STAGE dbt_integration_test")
        project.run_sql(
            f"PUT file://{project.project_root}/seeds/iris.csv @dbt_integration_test/;"
        )
        run_dbt(["run"])


# https://github.com/dbt-labs/dbt-snowflake/issues/393 is notorious for being triggered on some
# environments but not others. As of writing this, we don't know the true root cause. This test may
# not fail on all systems with problems regarding custom schema model configurations.
class TestCustomSchemaWorks:
    @pytest.fixture(scope="class")
    def models(self):
        return {"custom_target_model.py": models__custom_target_model}

    @pytest.fixture(scope="function", autouse=True)
    def teardown_method(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=f"{project.test_schema}_MY_CUSTOM_SCHEMA"
            )
            project.adapter.drop_schema(relation)

    def test_custom_target(self, project):
        results = run_dbt()
        assert results[0].node.schema == f"{project.test_schema}_MY_CUSTOM_SCHEMA"


EXTERNAL_ACCESS_INTEGRATION_MODE = """
import pandas
import snowflake.snowpark as snowpark

def model(dbt, session: snowpark.Session):
    dbt.config(
        materialized="table",
        external_access_integrations=["test_external_access_integration"],
        packages=["httpx==0.26.0"]
    )
    import httpx
    return session.create_dataframe(
        pandas.DataFrame(
            [{"result": httpx.get(url="https://www.google.com").status_code}]
        )
    )
"""


class TestExternalAccessIntegration:
    @pytest.fixture(scope="class")
    def models(self):
        return {"external_access_integration_python_model.py": EXTERNAL_ACCESS_INTEGRATION_MODE}

    def test_external_access_integration(self, project):
        project.run_sql(
            "create or replace network rule test_network_rule type = host_port mode = egress value_list= ('www.google.com:443');"
        )
        project.run_sql(
            "create or replace external access integration test_external_access_integration allowed_network_rules = (test_network_rule) enabled = true;"
        )
        run_dbt(["run"])


TEST_RUN_ID = uuid.uuid4().hex
TEST_SECRET = f"test_secret_{TEST_RUN_ID}"
TEST_NETWORK_RULE = f"test_network_rule_{TEST_RUN_ID}"
TEST_EXTERNAL_ACCESS_INTEGRATION = f"test_external_access_integration_{TEST_RUN_ID}"
SECRETS_MODE = f"""
import pandas
import snowflake.snowpark as snowpark

def model(dbt, session: snowpark.Session):
    dbt.config(
        materialized="table",
        secrets={{"secret_variable_name": "{TEST_SECRET}"}},
        external_access_integrations=["{TEST_EXTERNAL_ACCESS_INTEGRATION}"],
    )
    import _snowflake
    return session.create_dataframe(
        pandas.DataFrame(
            [{{"secret_value": _snowflake.get_generic_secret_string('secret_variable_name')}}]
        )
    )
"""


class TestSecrets:
    @pytest.fixture(scope="class")
    def models(self):
        return {"secret_python_model.py": SECRETS_MODE}

    @pytest.fixture(scope="class")
    def profiles_config_update(self):
        return {"retry_all": True, "connect_retries": 3}

    def test_secrets(self, project):
        project.run_sql(
            f"create or replace secret {TEST_SECRET} type = generic_string secret_string='secret value';"
        )

        project.run_sql(
            f"create or replace network rule {TEST_NETWORK_RULE} type = host_port mode = egress value_list= ('www.google.com:443');"
        )

        project.run_sql(
            f"create or replace external access integration {TEST_EXTERNAL_ACCESS_INTEGRATION} "
            f"allowed_network_rules = ({TEST_NETWORK_RULE}) "
            f"allowed_authentication_secrets = ({TEST_SECRET}) enabled = true;"
        )

        run_dbt(["run"])

        project.run_sql(f"drop secret if exists {TEST_SECRET};")
        project.run_sql(f"drop network rule if exists {TEST_NETWORK_RULE};")
        project.run_sql(
            f"drop external access integration if exists {TEST_EXTERNAL_ACCESS_INTEGRATION};"
        )


class TestEmptyModeWithPythonModel(BasePythonEmptyTests):
    pass


@pytest.mark.skip(reason="Sample mode is unsupported on 1.9")
class TestSampleModeWithPythonModel(BasePythonSampleTests):
    pass



import pytest
from dbt.tests.util import check_relations_equal, run_dbt
from tests.functional.adapter.custom_schema_tests.seeds import seed_agg_csv, seed_csv

_VIEW_1_SQL = """
select * from {{ ref('seed') }}
""".lstrip()

_VIEW_2_SQL = """
{{ config(schema='custom') }}

select * from {{ ref('view_1') }}
""".lstrip()

_VIEW_3_SQL = """
{{ config(schema='test', materialized='table') }}


with v1 as (

    select * from {{ ref('view_1') }}

),

v2 as (

    select * from {{ ref('view_2') }}

),

combined as (

    select last_name from v1
    union all
    select last_name from v2

)

select
    last_name,
    count(*) as count

from combined
group by 1
""".lstrip()


class TestCustomProjectSchemaWithPrefix:
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": seed_csv, "agg.csv": seed_agg_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {"view_1.sql": _VIEW_1_SQL, "view_2.sql": _VIEW_2_SQL, "view_3.sql": _VIEW_3_SQL}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"schema": "dbt_test"}}

    @pytest.fixture(scope="function")
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            alt_schema_list = ["DBT_TEST", "CUSTOM", "TEST"]
            for alt_schema in alt_schema_list:
                alt_test_schema = f"{project.test_schema}_{alt_schema}"
                relation = project.adapter.Relation.create(
                    database=project.database, schema=alt_test_schema
                )
                project.adapter.drop_schema(relation)

    def test__snowflake__custom_schema_with_prefix(self, project, clean_up):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 2
        results = run_dbt()
        assert len(results) == 3

        db_with_schema = f"{project.database}.{project.test_schema}"
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.SEED", f"{db_with_schema}_DBT_TEST.VIEW_1"]
        )
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.SEED", f"{db_with_schema}_CUSTOM.VIEW_2"]
        )
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.AGG", f"{db_with_schema}_TEST.VIEW_3"]
        )



import pytest
from dbt.tests.adapter.grants.test_incremental_grants import BaseIncrementalGrants
from dbt.tests.adapter.grants.test_invalid_grants import BaseInvalidGrants
from dbt.tests.adapter.grants.test_model_grants import BaseModelGrants
from dbt.tests.adapter.grants.test_seed_grants import BaseSeedGrants
from dbt.tests.adapter.grants.test_snapshot_grants import BaseSnapshotGrants


class BaseCopyGrantsSnowflake:
    # Try every test case without copy_grants enabled (default),
    # and with copy_grants enabled (this base class)
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "+copy_grants": True,
            },
            "seeds": {
                "+copy_grants": True,
            },
            "snapshots": {
                "+copy_grants": True,
            },
        }


class TestInvalidGrantsSnowflake(BaseInvalidGrants):
    def grantee_does_not_exist_error(self):
        return "does not exist or not authorized"

    def privilege_does_not_exist_error(self):
        return "unexpected"


class TestModelGrantsSnowflake(BaseModelGrants):
    pass


class TestModelGrantsCopyGrantsSnowflake(BaseCopyGrantsSnowflake, BaseModelGrants):
    pass


class TestIncrementalGrantsSnowflake(BaseIncrementalGrants):
    pass


class TestIncrementalCopyGrantsSnowflake(BaseCopyGrantsSnowflake, BaseIncrementalGrants):
    pass


class TestSeedGrantsSnowflake(BaseSeedGrants):
    pass


class TestSeedCopyGrantsSnowflake(BaseCopyGrantsSnowflake, BaseSeedGrants):
    pass


class TestSnapshotGrants(BaseSnapshotGrants):
    pass


class TestSnapshotCopyGrantsSnowflake(BaseCopyGrantsSnowflake, BaseSnapshotGrants):
    pass



import pytest

from dbt.tests.adapter.functions.test_udafs import BasicPythonUDAF, PythonUDAFDefaultArgSupport


class TestSnowflakePythonUDAF(BasicPythonUDAF):
    pass


class TestSnowflakePythonUDAFDefaultArgSupport(PythonUDAFDefaultArgSupport):
    expect_default_arg_support = True


class TestSnowflakePythonUDAFVolatilitySupport(BasicPythonUDAF):
    def check_function_volatility(self, sql: str):
        assert "VOLATILE" in sql

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "functions": {"+volatility": "non-deterministic"},
        }



import pytest
import os
from dbt.tests.util import check_relations_equal, check_table_does_exist, run_dbt
from tests.functional.adapter.custom_schema_tests.seeds import seed_agg_csv, seed_csv

_VIEW_1_SQL = """
select * from {{ ref('seed') }}
""".lstrip()

_VIEW_2_SQL = """
{{ config(database='alt') }}
select * from {{ ref('view_1') }}
""".lstrip()

_VIEW_3_SQL = """
{{ config(database='alt', materialized='table') }}


with v1 as (

    select * from {{ ref('view_1') }}

),

v2 as (

    select * from {{ ref('view_2') }}

),

combined as (

    select last_name from v1
    union all
    select last_name from v2

)

select
    last_name,
    count(*) as count

from combined
group by 1
""".lstrip()

_CUSTOM_DB_SQL = """
{% macro generate_database_name(database_name, node) %}
    {% if database_name == 'alt' %}
        {{ env_var('SNOWFLAKE_TEST_ALT_DATABASE') }}
    {% elif database_name %}
        {{ database_name }}
    {% else %}
        {{ target.database }}
    {% endif %}
{% endmacro %}
""".lstrip()

ALT_DATABASE = os.getenv("SNOWFLAKE_TEST_ALT_DATABASE")


class TestOverrideDatabase:
    @pytest.fixture(scope="class")
    def macros(self):
        return {
            "custom_db.sql": _CUSTOM_DB_SQL,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": seed_csv, "agg.csv": seed_agg_csv}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "view_1.sql": _VIEW_1_SQL,
            "view_2.sql": _VIEW_2_SQL,
            "view_3.sql": _VIEW_3_SQL,
        }

    @pytest.fixture(scope="function")
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=ALT_DATABASE, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    def test_snowflake_override_generate_db_name(self, project, clean_up):
        seed_results = run_dbt(["seed", "--full-refresh"])
        assert len(seed_results) == 2

        db_with_schema = f"{project.database}.{project.test_schema}"
        alt_db_with_schema = f"{ALT_DATABASE}.{project.test_schema}"
        seed_table = "SEED"
        agg_table = "AGG"
        view_1 = "VIEW_1"
        view_2 = "VIEW_2"
        view_3 = "VIEW_3"

        check_table_does_exist(project.adapter, f"{db_with_schema}.{seed_table}")
        check_table_does_exist(project.adapter, f"{db_with_schema}.{agg_table}")

        results = run_dbt()
        assert len(results) == 3

        check_table_does_exist(project.adapter, f"{db_with_schema}.{view_1}")
        check_table_does_exist(project.adapter, f"{alt_db_with_schema}.{view_2}")
        check_table_does_exist(project.adapter, f"{alt_db_with_schema}.{view_3}")

        # not overridden
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.{seed_table}", f"{db_with_schema}.{view_1}"]
        )

        # overridden
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.{seed_table}", f"{alt_db_with_schema}.{view_2}"]
        )
        check_relations_equal(
            project.adapter, [f"{db_with_schema}.{agg_table}", f"{alt_db_with_schema}.{view_3}"]
        )



MY_SEED = """
id,value
1,100
2,200
3,300
""".strip()


MY_TABLE = """
{{ config(
    materialized='table',
) }}
select * from {{ ref('my_seed') }}
"""


MY_VIEW = """
{{ config(
    materialized='view',
) }}
select * from {{ ref('my_seed') }}
"""


MY_DYNAMIC_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='30 minutes',
) }}
select * from {{ ref('my_seed') }}
"""



import pytest
from dbt.tests.adapter.query_comment.test_query_comment import (
    BaseQueryComments,
    BaseMacroQueryComments,
    BaseMacroArgsQueryComments,
    BaseMacroInvalidQueryComments,
    BaseNullQueryComments,
    BaseEmptyQueryComments,
)


class TestQueryCommentsSnowflake(BaseQueryComments):
    def test_matches_comment(self, project):
        logs = self.run_get_json()
        # No newline in logs because query comment is appended and newline stripped.
        assert r"/* dbt\nrules! */" in logs


class TestMacroQueryCommentsSnowflake(BaseMacroQueryComments):
    def test_matches_comment(self, project):
        logs = self.run_get_json()
        # No newline in logs because query comment is appended and newline stripped.
        assert r"/* dbt macros\nare pretty cool */" in logs


class TestMacroArgsQueryCommentsSnowflake(BaseMacroArgsQueryComments):
    @pytest.mark.skip(
        "This test is incorrectly comparing the version of `dbt-core`"
        "to the version of `dbt-snowflake`, which is not always the same."
    )
    def test_matches_comment(self, project, get_package_version):
        pass


class TestMacroInvalidQueryCommentsSnowflake(BaseMacroInvalidQueryComments):
    pass


class TestNullQueryCommentsSnowflake(BaseNullQueryComments):
    pass


class TestEmptyQueryCommentsSnowflake(BaseEmptyQueryComments):
    pass



SEED = """
id,value
1,100
2,200
3,300
""".strip()


DYNAMIC_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
    refresh_mode='INCREMENTAL',
) }}
select * from {{ ref('my_seed') }}
"""


EXPLICIT_AUTO_DYNAMIC_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
    refresh_mode='AUTO',
) }}
select * from {{ ref('my_seed') }}
"""

IMPLICIT_AUTO_DYNAMIC_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_TABLE_DOWNSTREAM = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='DOWNSTREAM',
    refresh_mode='INCREMENTAL',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_ICEBERG_TABLE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
    refresh_mode='INCREMENTAL',
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_TABLE_ALTER = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='5 minutes',
    refresh_mode='INCREMENTAL',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_TABLE_REPLACE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
    refresh_mode='FULL',
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_ICEBERG_TABLE_ALTER = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='5 minutes',
    refresh_mode='INCREMENTAL',
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
) }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_ICEBERG_TABLE_REPLACE = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='2 minutes',
    refresh_mode='FULL',
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
    base_location_subpath="subpath",
) }}
select * from {{ ref('my_seed') }}
"""

DYNAMIC_TABLE_CUSTOM_SCHEMA = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='30 minutes',
    schema='custom_schema'
) }}
select * from {{ ref('simple_model') }}
"""

DYNAMIC_TABLE_CUSTOM_DB_SCHEMA = """
{{ config(
    materialized='dynamic_table',
    snowflake_warehouse='DBT_TESTING',
    target_lag='30 minutes',
    schema='custom_schema'
) }}
select * from {{ ref('simple_model') }}
"""

SIMPLE_MODEL = """
{{ config(
    materialized='table'
) }}
SELECT 1 as id
"""



import dataclasses
from typing import Optional

from dbt_common.record import Record, Recorder


@dataclasses.dataclass
class CursorGetSqlStateParams:
    connection_name: str


@dataclasses.dataclass
class CursorGetSqlStateResult:
    msg: Optional[str]


@Recorder.register_record_type
class CursorGetSqlStateRecord(Record):
    params_cls = CursorGetSqlStateParams
    result_cls = CursorGetSqlStateResult
    group = "Database"



import json
import os

from dbt.tests.util import run_dbt

from dbt.tests.adapter.persist_docs.test_persist_docs import (
    BasePersistDocs,
    BasePersistDocsColumnMissing,
    BasePersistDocsCommentOnQuotedColumn,
)


class TestPersistDocs(BasePersistDocs):
    def _assert_common_comments(self, *comments):
        for comment in comments:
            assert '"with double quotes"' in comment
            assert """'''abc123'''""" in comment
            assert "\n" in comment
            assert (
                "Some [$]lbl[$] labeled [$]lbl[$] and [$][$] unlabeled [$][$] dollar-quoting"
                in comment
            )
            assert "/* comment */" in comment
            if os.name == "nt":
                assert "--\r\n" in comment or "--\n" in comment
            else:
                assert "--\n" in comment

    def _assert_has_table_comments(self, table_node):
        table_comment = table_node["metadata"]["comment"]
        assert table_comment.startswith("Table model description")

        table_id_comment = table_node["columns"]["ID"]["comment"]
        assert table_id_comment.startswith("id Column description")

        table_name_comment = table_node["columns"]["NAME"]["comment"]
        assert table_name_comment.startswith("Some stuff here and then a call to")

        self._assert_common_comments(table_comment, table_id_comment, table_name_comment)

    def _assert_has_view_comments(
        self, view_node, has_node_comments=True, has_column_comments=True
    ):
        view_comment = view_node["metadata"]["comment"]
        if has_node_comments:
            assert view_comment.startswith("View model description")
            self._assert_common_comments(view_comment)
        else:
            assert not view_comment

        view_id_comment = view_node["columns"]["ID"]["comment"]
        if has_column_comments:
            assert view_id_comment.startswith("id Column description")
            self._assert_common_comments(view_id_comment)
        else:
            assert not view_id_comment

        view_name_comment = view_node["columns"]["NAME"]["comment"]
        assert not view_name_comment


class TestPersistDocsColumnMissing(BasePersistDocsColumnMissing):
    def test_missing_column(self, project):
        run_dbt(["docs", "generate"])
        with open("target/catalog.json") as fp:
            catalog_data = json.load(fp)
        assert "nodes" in catalog_data

        table_node = catalog_data["nodes"]["model.test.missing_column"]
        table_id_comment = table_node["columns"]["ID"]["comment"]
        assert table_id_comment.startswith("test id column description")


class TestPersistDocsCommentOnQuotedColumn(BasePersistDocsCommentOnQuotedColumn):
    pass



from copy import deepcopy
from types import SimpleNamespace

import pytest

from dbt.adapters.snowflake import constants
from dbt.adapters.snowflake.catalogs import BuiltInCatalogIntegration


@pytest.fixture
def fake_integration() -> BuiltInCatalogIntegration:
    return BuiltInCatalogIntegration(constants.DEFAULT_BUILT_IN_CATALOG)


model_base = SimpleNamespace(
    database="my_database",
    schema="my_schema",
    identifier="my_table",
    config={
        "catalog": "snowflake",
        "external_volume": "s3_iceberg_snow",
    },
)


@pytest.mark.parametrize(
    "config,expected",
    [
        (
            {},
            "_dbt/my_schema/my_table",
        ),
        (
            {"base_location_root": None, "base_location_subpath": None},
            "_dbt/my_schema/my_table",
        ),
        (
            {"base_location_root": "root_path", "base_location_subpath": "subpath"},
            "root_path/my_schema/my_table/subpath",
        ),
        (
            {"base_location_subpath": "subpath"},
            "_dbt/my_schema/my_table/subpath",
        ),
        (
            {"base_location_root": "root_path"},
            "root_path/my_schema/my_table",
        ),
    ],
)
def test_iceberg_base_location_built_in(fake_integration, config, expected):
    """Test when base_location_root and base_location_subpath are provided"""
    model = deepcopy(model_base)
    model.config.update(config)
    relation = fake_integration.build_relation(model)
    assert relation.base_location == expected


@pytest.mark.parametrize(
    "config,expected",
    [
        (None, None),
        (False, "FALSE"),
        (True, "TRUE"),
        ("False", "FALSE"),
        ("True", "TRUE"),
    ],
)
def test_change_tracking_model_config(fake_integration, config, expected):
    model = deepcopy(model_base)
    model.config.update({"change_tracking": config})
    relation = fake_integration.build_relation(model)
    assert relation.change_tracking == expected


@pytest.mark.parametrize(
    "user_input",
    [
        "0",
        "",
        "None",
    ],
)
def test_change_tracking_invalid_model_config(fake_integration, user_input):
    model = deepcopy(model_base)
    model.config.update({"change_tracking": user_input})
    with pytest.raises(ValueError) as e:
        fake_integration.build_relation(model)
    assert "Invalid value for change_tracking" in str(e.value)



import os
from typing import List

import pytest

from pathlib import Path

from dbt.adapters.factory import get_adapter_by_type
from dbt.adapters.snowflake import SnowflakeRelation

from dbt.tests.util import run_dbt, get_connection


SEED = """
id,value
0,red
1,yellow
2,blue
""".strip()


VIEW = """
select * from {{ ref('my_seed') }}
"""


TABLE = """
{{ config(materialized='table') }}
select * from {{ ref('my_seed') }}
"""


DYNAMIC_TABLE = (
    """
{{ config(
    materialized='dynamic_table',
    target_lag='1 day',
    snowflake_warehouse='"""
    + os.getenv("SNOWFLAKE_TEST_WAREHOUSE", "")
    + """',
) }}
select * from {{ ref('my_seed') }}
"""
)

_MODEL_ICEBERG = """
{{
  config(
    materialized = "table",
    table_format="iceberg",
    external_volume="s3_iceberg_snow",
  )
}}

select 1 as id
"""


class ShowObjectsBase:
    @staticmethod
    def list_relations_without_caching(project) -> List[SnowflakeRelation]:
        my_adapter = get_adapter_by_type("snowflake")
        schema = my_adapter.Relation.create(
            database=project.database, schema=project.test_schema, identifier=""
        )
        with get_connection(my_adapter):
            relations = my_adapter.list_relations_without_caching(schema)
        return relations


class TestShowObjects(ShowObjectsBase):
    views: int = 10
    tables: int = 10
    dynamic_tables: int = 10

    @pytest.fixture(scope="class")
    def seeds(self):
        yield {"my_seed.csv": SEED}

    @pytest.fixture(scope="class")
    def models(self):
        models = {}
        models.update({f"my_view_{i}.sql": VIEW for i in range(self.views)})
        models.update({f"my_table_{i}.sql": TABLE for i in range(self.tables)})
        models.update(
            {f"my_dynamic_table_{i}.sql": DYNAMIC_TABLE for i in range(self.dynamic_tables)}
        )
        yield models

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])

    def test_list_relations_without_caching(self, project):
        relations = self.list_relations_without_caching(project)
        assert len([relation for relation in relations if relation.is_view]) == self.views
        assert (
            len([relation for relation in relations if relation.is_table])
            == self.tables + 1  # add the seed
        )
        assert (
            len([relation for relation in relations if relation.is_dynamic_table])
            == self.dynamic_tables
        )


class TestShowIcebergObjects(ShowObjectsBase):

    @pytest.fixture(scope="class")
    def models(self):
        return {"my_model.sql": _MODEL_ICEBERG}

    def test_quoting_ignore_flag_doesnt_break_iceberg_metadata(self, project):
        """https://github.com/dbt-labs/dbt-snowflake/issues/1227

        The list relations function involves a metadata sub-query. Regardless of
        QUOTED_IDENTIFIERS_IGNORE_CASE, this function will fail without proper
        normalization within the encapsulating python function after the macro invocation
        returns. This test verifies that normalization is working.
        """
        run_dbt(["run"])

        self.list_relations_without_caching(project)



import pytest

from pathlib import Path

from dbt.tests.util import run_dbt, rm_file, write_file

from tests.functional.iceberg import models


class TestIcebergTableBuilds:

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "first_table.sql": models._MODEL_BASIC_TABLE_MODEL,
            "first_table_literals.sql": models._MODEL_BASIC_TABLE_LITERALS,
            "iceberg_table.sql": models._MODEL_BASIC_ICEBERG_MODEL,
            "iceberg_builtin_table.sql": models._MODEL_BASIC_ICEBERG_BUILTIN_MODEL,
            "iceberg_tableb.sql": models._MODEL_BASIC_ICEBERG_MODEL_WITH_PATH,
            "iceberg_tablec.sql": models._MODEL_BASIC_ICEBERG_MODEL_WITH_PATH_SUBPATH,
            "table_built_on_iceberg_table.sql": models._MODEL_BUILT_ON_ICEBERG_TABLE,
            "dynamic_table.sql": models._MODEL_BASIC_DYNAMIC_TABLE_MODEL,
            "dynamic_table_builtin.sql": models._MODEL_BASIC_DYNAMIC_TABLE_ICEBERG_MODEL,
            "dynamic_tableb.sql": models._MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_PATH,
            "dynamic_tablec.sql": models._MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_PATH_SUBPATH,
            "dynamic_tabled.sql": models._MODEL_BASIC_DYNAMIC_TABLE_MODEL_WITH_SUBPATH,
        }

    def test_iceberg_tables_build_and_can_be_referred(self, project):
        run_results = run_dbt()
        assert len(run_results) == 12


class TestIcebergTableTypeBuildsOnExistingTable:

    @pytest.mark.parametrize(
        "start_model", [models._MODEL_TABLE_BEFORE_SWAP, models._MODEL_VIEW_BEFORE_SWAP]
    )
    def test_changing_model_types(self, project, start_model):
        model_file = project.project_root / Path("models") / Path("my_model.sql")

        write_file(start_model, model_file)
        run_results = run_dbt()
        assert len(run_results) == 1

        rm_file(model_file)
        write_file(models._MODEL_TABLE_FOR_SWAP_ICEBERG, model_file)
        run_results = run_dbt()
        assert len(run_results) == 1

        rm_file(model_file)
        write_file(start_model, model_file)
        run_results = run_dbt()
        assert len(run_results) == 1



from dataclasses import dataclass
from itertools import product
from typing import Optional

from dbt.tests.util import run_dbt
import pytest

from tests.functional.relation_tests import models
from tests.functional.utils import describe_dynamic_table, query_relation_type, update_model


@dataclass
class Model:
    model: str
    relation_type: str
    table_format: Optional[str] = "default"
    is_incremental: Optional[bool] = False

    @property
    def name(self) -> str:
        if self.is_incremental:
            name = f"{self.relation_type}_{self.table_format}_incremental"
        else:
            name = f"{self.relation_type}_{self.table_format}"
        return name

    @property
    def is_standard_table(self) -> bool:
        return self.relation_type == "table" and not self.is_incremental


@dataclass
class Scenario:
    initial: Model
    final: Model

    @property
    def name(self) -> str:
        return f"REPLACE_{self.initial.name}__WITH_{self.final.name}"

    @property
    def error_message(self) -> str:
        return f"Failed when migrating from: {self.initial.name} to: {self.final.name}"


relations = [
    Model(models.VIEW, "view"),
    Model(models.TABLE, "table", "default"),
    Model(models.INCREMENTAL_TABLE, "table", "default", is_incremental=True),
    Model(models.DYNAMIC_TABLE, "dynamic_table", "default"),
    Model(models.ICEBERG_TABLE, "table", "iceberg"),
    Model(models.INCREMENTAL_ICEBERG_TABLE, "table", "iceberg", is_incremental=True),
    Model(models.DYNAMIC_ICEBERG_TABLE, "dynamic_table", "iceberg"),
]
scenarios = [Scenario(*scenario) for scenario in product(relations, relations)]


def requires_full_refresh(scenario) -> bool:
    return any(
        [
            # we can only swap incremental to table and back if both are iceberg
            scenario.initial.is_incremental
            and scenario.final.is_standard_table
            and scenario.initial.table_format != scenario.final.table_format,
            scenario.initial.is_standard_table
            and scenario.final.is_incremental
            and scenario.initial.table_format != scenario.final.table_format,
            # we can't swap from an incremental to a dynamic table because the materialization does not handle this case
            scenario.initial.relation_type == "dynamic_table" and scenario.final.is_incremental,
        ]
    )


class TestRelationTypeChange:

    @staticmethod
    def include(scenario) -> bool:
        return not requires_full_refresh(scenario)

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        return {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        yield {
            f"{scenario.name}.sql": scenario.initial.model
            for scenario in scenarios
            if self.include(scenario)
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        for scenario in scenarios:
            if self.include(scenario):
                update_model(project, scenario.name, scenario.final.model)
        # allow for dbt to fail so that we can see which scenarios pass and which scenarios fail
        try:
            run_dbt(["run"], expect_pass=False)
        except Exception:
            pass

    @pytest.mark.parametrize("scenario", scenarios, ids=[scenario.name for scenario in scenarios])
    def test_replace(self, project, scenario):
        if self.include(scenario):
            relation_type = query_relation_type(project, scenario.name)
            assert relation_type == scenario.final.relation_type, scenario.error_message
        else:
            pytest.skip()


class TestRelationTypeChangeFullRefreshRequired(TestRelationTypeChange):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"full_refresh": True}}

    @staticmethod
    def include(scenario) -> bool:
        return requires_full_refresh(scenario)



import unittest
from unittest import mock
import re
from jinja2 import Environment, FileSystemLoader


class TestSnowflakeAlterRelationCommentMacro(unittest.TestCase):
    def setUp(self):
        self.jinja_env = Environment(
            loader=FileSystemLoader("src/dbt/include/snowflake/macros"),
            extensions=[
                "jinja2.ext.do",
            ],
        )

        self.config = {}
        self.default_context = {
            "validation": mock.Mock(),
            "model": mock.Mock(),
            "exceptions": mock.Mock(),
            "config": mock.Mock(),
            "adapter": mock.Mock(),
            "return": lambda r: r,
        }
        self.default_context["config"].get = lambda key, default=None, **kwargs: self.config.get(
            key, default
        )

    def __get_template(self, template_filename):
        return self.jinja_env.get_template(template_filename, globals=self.default_context)

    def __run_macro(self, template, name, relation, relation_comment):
        def dispatch(macro_name, macro_namespace=None, packages=None):
            return getattr(template.module, f"snowflake__{macro_name}")

        self.default_context["adapter"].dispatch = dispatch

        value = getattr(template.module, name)(relation, relation_comment)
        return re.sub(r"\s+", " ", value.strip())

    def __create_mock_relation(
        self,
        relation_type="table",
        is_dynamic_table=False,
        is_iceberg_format=False,
        database="test_db",
        schema="test_schema",
        identifier="test_table",
    ):
        """Create a mock relation object with the specified properties"""
        mock_relation = mock.Mock()
        mock_relation.type = relation_type
        mock_relation.is_dynamic_table = is_dynamic_table
        mock_relation.is_iceberg_format = is_iceberg_format
        mock_relation.render.return_value = f"{database}.{schema}.{identifier}"
        return mock_relation

    def test_macros_load(self):
        """Test that the adapters.sql template loads without errors"""
        self.jinja_env.get_template("adapters.sql")

    def test_alter_relation_comment_regular_table(self):
        """Test alter_relation_comment for a regular table"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table")
        comment = "This is a test comment"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "comment on table test_db.test_schema.test_table IS $$This is a test comment$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_regular_view(self):
        """Test alter_relation_comment for a regular view"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="view")
        comment = "This is a view comment"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "comment on view test_db.test_schema.test_table IS $$This is a view comment$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_dynamic_table(self):
        """Test alter_relation_comment for a dynamic table"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table", is_dynamic_table=True)
        comment = "This is a dynamic table comment"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "comment on dynamic table test_db.test_schema.test_table IS $$This is a dynamic table comment$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_iceberg_table(self):
        """Test alter_relation_comment for an iceberg table"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table", is_iceberg_format=True)
        comment = "This is an iceberg table comment"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "alter iceberg table test_db.test_schema.test_table set comment = $$This is an iceberg table comment$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_dollar_sign_escaping(self):
        """Test that dollar signs in comments are properly escaped"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table")
        comment = "This comment has a $dollar sign and $multiple $signs"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "comment on table test_db.test_schema.test_table IS $$This comment has a [$]dollar sign and [$]multiple [$]signs$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_iceberg_table_dollar_sign_escaping(self):
        """Test that dollar signs in iceberg table comments are properly escaped"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table", is_iceberg_format=True)
        comment = "Iceberg table with $dollar signs: $100 and $200"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "alter iceberg table test_db.test_schema.test_table set comment = $$Iceberg table with [$]dollar signs: [$]100 and [$]200$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_empty_comment(self):
        """Test alter_relation_comment with an empty comment"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table")
        comment = ""

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        expected = "comment on table test_db.test_schema.test_table IS $$$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_multiline_comment(self):
        """Test alter_relation_comment with a multiline comment"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="view")
        comment = """This is a multiline
        comment with
        multiple lines"""

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        # Note: The macro should preserve the multiline structure
        expected_comment = comment.replace("$", "[$]")
        expected = f"comment on view test_db.test_schema.test_table IS $${expected_comment}$$;"
        # Normalize whitespace for comparison since our __run_macro helper compresses whitespace
        self.assertTrue("multiline" in sql and "multiple lines" in sql)
        self.assertTrue(sql.startswith("comment on view"))
        self.assertTrue(sql.endswith("$$;"))

    def test_alter_relation_comment_special_characters(self):
        """Test alter_relation_comment with various special characters"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(relation_type="table")
        comment = (
            "Comment with 'quotes', \"double quotes\", and other chars: @#%^&*()[]{}|\\;:,.<>?"
        )

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        # Only dollar signs should be escaped, other characters should remain as-is
        expected = "comment on table test_db.test_schema.test_table IS $$Comment with 'quotes', \"double quotes\", and other chars: @#%^&*()[]{}|\\;:,.<>?$$;"
        self.assertEqual(sql, expected)

    def test_alter_relation_comment_iceberg_dynamic_precedence(self):
        """Test that iceberg format takes precedence over dynamic table"""
        template = self.__get_template("adapters.sql")
        relation = self.__create_mock_relation(
            relation_type="table", is_dynamic_table=True, is_iceberg_format=True
        )
        comment = "This is both dynamic and iceberg"

        sql = self.__run_macro(template, "snowflake__alter_relation_comment", relation, comment)

        # Should use iceberg syntax, not dynamic table syntax
        expected = "alter iceberg table test_db.test_schema.test_table set comment = $$This is both dynamic and iceberg$$;"
        self.assertEqual(sql, expected)



from tests.functional.relation_tests.base import RelationOperation


class TestView(RelationOperation):

    def test_get_create_backup_and_rename_intermediate_sql(self, project):
        args = {
            "database": project.database,
            "schema": project.test_schema,
            "identifier": "my_view",
            "relation_type": "view",
        }
        expected_statement = (
            f"alter view {project.database}.{project.test_schema}.my_view "
            f"rename to {project.database}.{project.test_schema}.my_view__dbt_backup"
        )
        self.assert_operation(project, "test__get_create_backup_sql", args, expected_statement)

        expected_statement = (
            f"alter view {project.database}.{project.test_schema}.my_view__dbt_tmp "
            f"rename to {project.database}.{project.test_schema}.my_view"
        )
        self.assert_operation(
            project, "test__get_rename_intermediate_sql", args, expected_statement
        )



import pytest

from dbt.tests.util import run_dbt, run_dbt_and_capture

from tests.functional.generic_test_tests import _files


class TestWarehouseConfig:

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "colors.csv": _files.SEED__COLORS,
            "facts.csv": _files.SEED__FACTS,
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup(self, project):
        run_dbt(["seed"])
        run_dbt(["run"])
        yield


class TestWarehouseConfigControl(TestWarehouseConfig):

    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": _files.SCHEMA__CONTROL}

    def test_expected_warehouse(self, project):
        results, logs = run_dbt_and_capture(["test"])
        assert len(results) == 1


class TestWarehouseConfigExplicitWarehouse(TestWarehouseConfig):

    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": _files.SCHEMA__EXPLICIT_WAREHOUSE}

    def test_expected_warehouse(self, project):
        _, logs = run_dbt_and_capture(["test", "--log-level", "debug"])
        assert "use warehouse " in logs


class TestWarehouseConfigNotNull(TestWarehouseConfig):

    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": _files.SCHEMA__NOT_NULL}

    def test_expected_warehouse(self, project):
        _, logs = run_dbt_and_capture(["test", "--log-level", "debug"], expect_pass=False)
        assert "use warehouse " in logs



import pytest

from dbt.tests.util import run_dbt

from tests.functional.relation_tests.dynamic_table_tests import models
from tests.functional.utils import describe_dynamic_table, update_model


class Changes:
    iceberg: bool = False

    @pytest.fixture(scope="class", autouse=True)
    def seeds(self):
        yield {"my_seed.csv": models.SEED}

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        my_models = {
            "dynamic_table_alter.sql": models.DYNAMIC_TABLE,
            "dynamic_table_replace.sql": models.DYNAMIC_TABLE,
        }
        if self.iceberg:
            my_models.update(
                {
                    "dynamic_table_iceberg_alter.sql": models.DYNAMIC_ICEBERG_TABLE,
                    "dynamic_table_iceberg_replace.sql": models.DYNAMIC_ICEBERG_TABLE,
                }
            )
        yield my_models

    @pytest.fixture(scope="function", autouse=True)
    def setup_class(self, project):
        run_dbt(["seed"])
        yield
        project.run_sql(f"drop schema if exists {project.test_schema} cascade")

    @pytest.fixture(scope="function", autouse=True)
    def setup_method(self, project, setup_class):
        # make sure the model in the data reflects the files each time
        run_dbt(["run", "--full-refresh"])
        self.assert_changes_are_not_applied(project)

        update_model(project, "dynamic_table_alter", models.DYNAMIC_TABLE_ALTER)
        update_model(project, "dynamic_table_replace", models.DYNAMIC_TABLE_REPLACE)
        if self.iceberg:
            update_model(
                project, "dynamic_table_iceberg_alter", models.DYNAMIC_ICEBERG_TABLE_ALTER
            )
            update_model(
                project, "dynamic_table_iceberg_replace", models.DYNAMIC_ICEBERG_TABLE_REPLACE
            )

        yield

        update_model(project, "dynamic_table_alter", models.DYNAMIC_TABLE)
        update_model(project, "dynamic_table_replace", models.DYNAMIC_TABLE)
        if self.iceberg:
            update_model(project, "dynamic_table_iceberg_alter", models.DYNAMIC_ICEBERG_TABLE)
            update_model(project, "dynamic_table_iceberg_replace", models.DYNAMIC_ICEBERG_TABLE)

    def assert_changes_are_applied(self, project):
        altered = describe_dynamic_table(project, "dynamic_table_alter")
        assert altered.snowflake_warehouse == "DBT_TESTING"
        assert altered.target_lag == "5 minutes"  # this updated
        assert altered.refresh_mode == "INCREMENTAL"

        replaced = describe_dynamic_table(project, "dynamic_table_replace")
        assert replaced.snowflake_warehouse == "DBT_TESTING"
        assert replaced.target_lag == "2 minutes"
        assert replaced.refresh_mode == "FULL"  # this updated

        if self.iceberg:
            altered_iceberg = describe_dynamic_table(project, "dynamic_table_iceberg_alter")
            assert altered_iceberg.snowflake_warehouse == "DBT_TESTING"
            assert altered_iceberg.target_lag == "5 minutes"  # this updated
            assert altered_iceberg.refresh_mode == "INCREMENTAL"

            replaced_iceberg = describe_dynamic_table(project, "dynamic_table_iceberg_replace")
            assert replaced_iceberg.snowflake_warehouse == "DBT_TESTING"
            assert replaced_iceberg.target_lag == "2 minutes"
            assert replaced_iceberg.refresh_mode == "FULL"  # this updated

    def assert_changes_are_not_applied(self, project):
        altered = describe_dynamic_table(project, "dynamic_table_alter")
        assert altered.snowflake_warehouse == "DBT_TESTING"
        assert altered.target_lag == "2 minutes"  # this would have updated, but didn't
        assert altered.refresh_mode == "INCREMENTAL"

        replaced = describe_dynamic_table(project, "dynamic_table_replace")
        assert replaced.snowflake_warehouse == "DBT_TESTING"
        assert replaced.target_lag == "2 minutes"
        assert replaced.refresh_mode == "INCREMENTAL"  # this would have updated, but didn't

        if self.iceberg:
            altered_iceberg = describe_dynamic_table(project, "dynamic_table_iceberg_alter")
            assert altered_iceberg.snowflake_warehouse == "DBT_TESTING"
            assert altered_iceberg.target_lag == "2 minutes"  # this would have updated, but didn't
            assert altered_iceberg.refresh_mode == "INCREMENTAL"

            replaced_iceberg = describe_dynamic_table(project, "dynamic_table_iceberg_replace")
            assert replaced_iceberg.snowflake_warehouse == "DBT_TESTING"
            assert replaced_iceberg.target_lag == "2 minutes"
            assert (
                replaced_iceberg.refresh_mode == "INCREMENTAL"
            )  # this would have updated, but didn't

    def test_full_refresh_is_always_successful(self, project):
        # this always passes and always changes the configuration, regardless of on_configuration_change
        # and regardless of whether the changes require a replace versus an alter
        run_dbt(["run", "--full-refresh"])
        self.assert_changes_are_applied(project)


class TestChangesApply(Changes):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"on_configuration_change": "apply"}}

    def test_changes_are_applied(self, project):
        # this passes and changes the configuration
        run_dbt(["run"])
        self.assert_changes_are_applied(project)


class TestChangesContinue(Changes):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"on_configuration_change": "continue"}}

    def test_changes_are_not_applied(self, project):
        # this passes but does not change the configuration
        run_dbt(["run"])
        self.assert_changes_are_not_applied(project)


class TestChangesFail(Changes):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {"models": {"on_configuration_change": "fail"}}

    def test_changes_are_not_applied(self, project):
        # this fails and does not change the configuration
        run_dbt(["run"], expect_pass=False)
        self.assert_changes_are_not_applied(project)



import pytest
from dbt.tests.util import run_dbt, check_relations_equal

import os

models__override_warehouse_sql = """
{{ config(snowflake_warehouse=env_var('SNOWFLAKE_TEST_ALT_WAREHOUSE', 'DBT_TEST_ALT'), materialized='table') }}
select current_warehouse() as warehouse
"""

models__expected_warehouse_sql = """
{{ config(materialized='table') }}
select '{{ env_var("SNOWFLAKE_TEST_ALT_WAREHOUSE", "DBT_TEST_ALT") }}' as warehouse
"""

models__invalid_warehouse_sql = """
{{ config(snowflake_warehouse='DBT_TEST_DOES_NOT_EXIST') }}
select current_warehouse() as warehouse
"""

project_config_models__override_warehouse_sql = """
{{ config(materialized='table') }}
select current_warehouse() as warehouse
"""

project_config_models__expected_warehouse_sql = """
{{ config(materialized='table') }}
select '{{ env_var("SNOWFLAKE_TEST_ALT_WAREHOUSE", "DBT_TEST_ALT") }}' as warehouse
"""

project_config_models__warehouse_sql = """
{{ config(materialized='table') }}
select current_warehouse() as warehouse
"""


class TestModelWarehouse:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "override_warehouse.sql": models__override_warehouse_sql,
            "expected_warehouse.sql": models__expected_warehouse_sql,
            "invalid_warehouse.sql": models__invalid_warehouse_sql,
        }

    def test_snowflake_override_ok(self, project):
        run_dbt(
            [
                "run",
                "--models",
                "override_warehouse",
                "expected_warehouse",
            ]
        )
        check_relations_equal(project.adapter, ["OVERRIDE_WAREHOUSE", "EXPECTED_WAREHOUSE"])

    def test_snowflake_override_noexist(self, project):
        run_dbt(["run", "--models", "invalid_warehouse"], expect_pass=False)


class TestConfigWarehouse:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "override_warehouse.sql": project_config_models__override_warehouse_sql,
            "expected_warehouse.sql": project_config_models__expected_warehouse_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "config-version": 2,
            "models": {
                "test": {
                    "snowflake_warehouse": os.getenv(
                        "SNOWFLAKE_TEST_ALT_WAREHOUSE", "DBT_TEST_ALT"
                    ),
                },
            },
        }

    def test_snowflake_override_ok(self, project):
        run_dbt(
            [
                "run",
                "--models",
                "override_warehouse",
                "expected_warehouse",
            ]
        )
        check_relations_equal(project.adapter, ["OVERRIDE_WAREHOUSE", "EXPECTED_WAREHOUSE"])


class TestInvalidConfigWarehouse:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "invalid_warehouse.sql": project_config_models__warehouse_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "config-version": 2,
            "models": {
                "test": {"snowflake_warehouse": "DBT_TEST_DOES_NOT_EXIST"},
            },
        }

    def test_snowflake_override_invalid(self, project):
        result = run_dbt(["run", "--models", "invalid_warehouse"], expect_pass=False)
        assert "Object does not exist, or operation cannot be performed" in result[0].message


class TestValidConfigWarehouse:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "valid_warehouse.sql": project_config_models__warehouse_sql,
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "config-version": 2,
            "models": {
                "test": {
                    "snowflake_warehouse": "DBT_TESTING",
                },
            },
        }

    def test_snowflake_warehouse_valid(self, project):
        result = run_dbt(["run", "--models", "valid_warehouse"])
        assert "DBT_TESTING" in result[0].node.config.get("snowflake_warehouse")



from dbt.tests.adapter.utils.test_array_append import BaseArrayAppend
from dbt.tests.adapter.utils.test_array_concat import BaseArrayConcat
from dbt.tests.adapter.utils.test_array_construct import BaseArrayConstruct
from dbt.tests.adapter.utils.test_any_value import BaseAnyValue
from dbt.tests.adapter.utils.test_bool_or import BaseBoolOr
from dbt.tests.adapter.utils.test_cast import BaseCast
from dbt.tests.adapter.utils.test_cast_bool_to_text import BaseCastBoolToText
from dbt.tests.adapter.utils.test_concat import BaseConcat
from dbt.tests.adapter.utils.test_current_timestamp import BaseCurrentTimestampAware
from dbt.tests.adapter.utils.test_date import BaseDate
from dbt.tests.adapter.utils.test_dateadd import BaseDateAdd
from dbt.tests.adapter.utils.test_datediff import BaseDateDiff
from dbt.tests.adapter.utils.test_date_spine import BaseDateSpine
from dbt.tests.adapter.utils.test_date_trunc import BaseDateTrunc
from dbt.tests.adapter.utils.test_escape_single_quotes import BaseEscapeSingleQuotesQuote
from dbt.tests.adapter.utils.test_escape_single_quotes import BaseEscapeSingleQuotesBackslash
from dbt.tests.adapter.utils.test_except import BaseExcept
from dbt.tests.adapter.utils.test_generate_series import BaseGenerateSeries
from dbt.tests.adapter.utils.test_get_intervals_between import BaseGetIntervalsBetween
from dbt.tests.adapter.utils.test_get_powers_of_two import BaseGetPowersOfTwo
from dbt.tests.adapter.utils.test_hash import BaseHash
from dbt.tests.adapter.utils.test_intersect import BaseIntersect
from dbt.tests.adapter.utils.test_last_day import BaseLastDay
from dbt.tests.adapter.utils.test_length import BaseLength
from dbt.tests.adapter.utils.test_listagg import BaseListagg
from dbt.tests.adapter.utils.test_position import BasePosition
from dbt.tests.adapter.utils.test_replace import BaseReplace
from dbt.tests.adapter.utils.test_right import BaseRight
from dbt.tests.adapter.utils.test_safe_cast import BaseSafeCast
from dbt.tests.adapter.utils.test_split_part import BaseSplitPart
from dbt.tests.adapter.utils.test_string_literal import BaseStringLiteral


class TestAnyValue(BaseAnyValue):
    pass


class TestArrayAppend(BaseArrayAppend):
    pass


class TestArrayConcat(BaseArrayConcat):
    pass


class TestArrayConstruct(BaseArrayConstruct):
    pass


class TestBoolOr(BaseBoolOr):
    pass


class TestCast(BaseCast):
    pass


class TestCastBoolToText(BaseCastBoolToText):
    pass


class TestConcat(BaseConcat):
    pass


# Use either BaseCurrentTimestampAware or BaseCurrentTimestampNaive but not both
class TestCurrentTimestamp(BaseCurrentTimestampAware):
    pass


class TestDate(BaseDate):
    pass


class TestDateAdd(BaseDateAdd):
    pass


class TestDateDiff(BaseDateDiff):
    pass


class TestDateSpine(BaseDateSpine):
    pass


class TestDateTrunc(BaseDateTrunc):
    pass


class TestEscapeSingleQuotesQuote(BaseEscapeSingleQuotesQuote):
    pass


class TestEscapeSingleQuotes(BaseEscapeSingleQuotesBackslash):
    pass


class TestExcept(BaseExcept):
    pass


class TestGenerateSeries(BaseGenerateSeries):
    pass


class TestGetIntervalsBeteween(BaseGetIntervalsBetween):
    pass


class TestGetPowersOfTwo(BaseGetPowersOfTwo):
    pass


class TestHash(BaseHash):
    pass


class TestIntersect(BaseIntersect):
    pass


class TestLastDay(BaseLastDay):
    pass


class TestLength(BaseLength):
    pass


class TestListagg(BaseListagg):
    pass


class TestPosition(BasePosition):
    pass


class TestReplace(BaseReplace):
    pass


class TestRight(BaseRight):
    pass


class TestSafeCast(BaseSafeCast):
    pass


class TestSplitPart(BaseSplitPart):
    pass


class TestStringLiteral(BaseStringLiteral):
    pass



import pytest

from dbt.tests.adapter.simple_seed.test_seed import SeedConfigBase
from dbt.tests.util import run_dbt


class TestSimpleBigSeedBatched(SeedConfigBase):
    @pytest.fixture(scope="class")
    def seeds(self):
        seed_data = ["seed_id"]
        seed_data.extend([str(i) for i in range(20_000)])
        return {"big_batched_seed.csv": "\n".join(seed_data)}

    def test_big_batched_seed(self, project):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 1



import pytest

from dbt.tests.adapter.basic.test_base import BaseSimpleMaterializations
from dbt.tests.adapter.basic.test_singular_tests import BaseSingularTests
from dbt.tests.adapter.basic.test_singular_tests_ephemeral import (
    BaseSingularTestsEphemeral,
)
from dbt.tests.adapter.basic.test_get_catalog_for_single_relation import (
    BaseGetCatalogForSingleRelation,
)
from dbt.tests.adapter.basic.test_empty import BaseEmpty
from dbt.tests.adapter.basic.test_ephemeral import BaseEphemeral
from dbt.tests.adapter.basic.test_incremental import BaseIncremental
from dbt.tests.adapter.basic.test_generic_tests import BaseGenericTests
from dbt.tests.adapter.basic.test_snapshot_check_cols import BaseSnapshotCheckCols
from dbt.tests.adapter.basic.test_snapshot_timestamp import BaseSnapshotTimestamp
from dbt.tests.adapter.basic.test_adapter_methods import BaseAdapterMethod
from dbt.tests.adapter.basic.test_docs_generate import BaseDocsGenerate
from dbt.tests.adapter.basic.expected_catalog import base_expected_catalog, no_stats
from dbt_common.contracts.metadata import CatalogTable, TableMetadata, ColumnMetadata, StatsItem

from dbt.adapters.snowflake.relation_configs import SnowflakeRelationType
from dbt.tests.util import run_dbt
from tests.functional.adapter.expected_stats import snowflake_stats


_MODEL_BASIC_TABLE_WITH_STRUCTURED_TYPES = """
{{ config(materialized='table') }}
select
    array_construct(1, 2, 3) as my_array,
    object_construct('languagepreference', 'english', 'localepreference', 'en_us') as my_object,
    { 'key1': 'value1', 'key2': 123, 'key3': true } AS my_object_literal,
    array_construct(0.1, 0.2, 0.3)::vector(float, 3) as my_vector
"""


class TestSimpleMaterializationsSnowflake(BaseSimpleMaterializations):
    pass


class TestSingularTestsSnowflake(BaseSingularTests):
    pass


class TestGetCatalogForSingleRelationSnowflake(BaseGetCatalogForSingleRelation):
    @pytest.fixture(scope="class")
    def current_role(self, project):
        return project.run_sql("select current_role()", fetch="one")[0]

    @pytest.fixture(scope="class")
    def expected_catalog_my_seed(self, project, current_role):
        return CatalogTable(
            metadata=TableMetadata(
                type=SnowflakeRelationType.Table.upper(),
                schema=project.test_schema.upper(),
                name="MY_SEED",
                database=project.database,
                comment="",
                owner=current_role,
            ),
            columns={
                "ID": ColumnMetadata(type="NUMBER", index=1, name="ID", comment=None),
                "FIRST_NAME": ColumnMetadata(
                    type="VARCHAR", index=2, name="FIRST_NAME", comment=None
                ),
                "EMAIL": ColumnMetadata(type="VARCHAR", index=3, name="EMAIL", comment=None),
                "IP_ADDRESS": ColumnMetadata(
                    type="VARCHAR", index=4, name="IP_ADDRESS", comment=None
                ),
                "UPDATED_AT": ColumnMetadata(
                    type="TIMESTAMP_NTZ", index=5, name="UPDATED_AT", comment=None
                ),
            },
            stats={
                "has_stats": StatsItem(
                    id="has_stats",
                    label="Has Stats?",
                    value=True,
                    include=False,
                    description="Indicates whether there are statistics for this table",
                ),
                "row_count": StatsItem(
                    id="row_count",
                    label="Row Count",
                    value=1,
                    include=True,
                    description="Number of rows in the table as reported by Snowflake",
                ),
                "bytes": StatsItem(
                    id="bytes",
                    label="Approximate Size",
                    value=2048,
                    include=True,
                    description="Size of the table as reported by Snowflake",
                ),
            },
            unique_id=None,
        )

    @pytest.fixture(scope="class")
    def expected_catalog_my_view_model(self, project, current_role):
        return CatalogTable(
            metadata=TableMetadata(
                type=SnowflakeRelationType.View.upper(),
                schema=project.test_schema.upper(),
                name="MY_VIEW_MODEL",
                database=project.database,
                comment="",
                owner=current_role,
            ),
            columns={
                "ID": ColumnMetadata(type="NUMBER", index=1, name="ID", comment=None),
                "FIRST_NAME": ColumnMetadata(
                    type="VARCHAR", index=2, name="FIRST_NAME", comment=None
                ),
                "EMAIL": ColumnMetadata(type="VARCHAR", index=3, name="EMAIL", comment=None),
                "IP_ADDRESS": ColumnMetadata(
                    type="VARCHAR", index=4, name="IP_ADDRESS", comment=None
                ),
                "UPDATED_AT": ColumnMetadata(
                    type="TIMESTAMP_NTZ", index=5, name="UPDATED_AT", comment=None
                ),
            },
            stats={
                "has_stats": StatsItem(
                    id="has_stats",
                    label="Has Stats?",
                    value=True,
                    include=False,
                    description="Indicates whether there are statistics for this table",
                ),
                "row_count": StatsItem(
                    id="row_count",
                    label="Row Count",
                    value=0,
                    include=False,
                    description="Number of rows in the table as reported by Snowflake",
                ),
                "bytes": StatsItem(
                    id="bytes",
                    label="Approximate Size",
                    value=0,
                    include=False,
                    description="Size of the table as reported by Snowflake",
                ),
            },
            unique_id=None,
        )

    @pytest.fixture(scope="class")
    def expected_catalog_my_table_model(self, project, current_role):
        return CatalogTable(
            metadata=TableMetadata(
                type=SnowflakeRelationType.Table.upper(),
                schema=project.test_schema.upper(),
                name="MY_TABLE_MODEL",
                database=project.database,
                comment="",
                owner=current_role,
            ),
            columns={
                "ID": ColumnMetadata(type="NUMBER", index=1, name="ID", comment=None),
                "FIRST_NAME": ColumnMetadata(
                    type="VARCHAR", index=2, name="FIRST_NAME", comment=None
                ),
                "EMAIL": ColumnMetadata(type="VARCHAR", index=3, name="EMAIL", comment=None),
                "IP_ADDRESS": ColumnMetadata(
                    type="VARCHAR", index=4, name="IP_ADDRESS", comment=None
                ),
                "UPDATED_AT": ColumnMetadata(
                    type="TIMESTAMP_NTZ", index=5, name="UPDATED_AT", comment=None
                ),
            },
            stats={
                "has_stats": StatsItem(
                    id="has_stats",
                    label="Has Stats?",
                    value=True,
                    include=False,
                    description="Indicates whether there are statistics for this table",
                ),
                "row_count": StatsItem(
                    id="row_count",
                    label="Row Count",
                    value=1,
                    include=True,
                    description="Number of rows in the table as reported by Snowflake",
                ),
                "bytes": StatsItem(
                    id="bytes",
                    label="Approximate Size",
                    value=2048,
                    include=True,
                    description="Size of the table as reported by Snowflake",
                ),
            },
            unique_id=None,
        )


class TestSingularTestsEphemeralSnowflake(BaseSingularTestsEphemeral):
    pass


class TestEmptySnowflake(BaseEmpty):
    pass


class TestEphemeralSnowflake(BaseEphemeral):
    pass


class TestIncrementalSnowflake(BaseIncremental):
    pass


class TestGenericTestsSnowflake(BaseGenericTests):
    pass


class TestSnapshotCheckColsSnowflake(BaseSnapshotCheckCols):
    pass


class TestSnapshotTimestampSnowflake(BaseSnapshotTimestamp):
    pass


class TestBaseAdapterMethodSnowflake(BaseAdapterMethod):
    @pytest.fixture(scope="class")
    def equal_tables(self):
        return ["MODEL", "EXPECTED"]


class TestDocsGenerateSnowflake(BaseDocsGenerate):
    @pytest.fixture(scope="class")
    def get_role(self, project):
        return project.run_sql("select current_role()", fetch="one")[0]

    @pytest.fixture(scope="class")
    def expected_catalog(self, project, get_role):
        return base_expected_catalog(
            project,
            role=get_role,
            id_type="NUMBER",
            text_type="TEXT",
            time_type="TIMESTAMP_NTZ",
            view_type="VIEW",
            table_type="BASE TABLE",
            model_stats=no_stats(),
            seed_stats=snowflake_stats(),
            case=lambda x: x.upper(),
            case_columns=False,
        )


class TestStructureTypesSnowflake:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "first_table.sql": _MODEL_BASIC_TABLE_WITH_STRUCTURED_TYPES,
        }

    def test_structured_types_parse_right(self, project):
        run_results = run_dbt()
        assert len(run_results) == 1
        run_results = run_dbt()
        assert len(run_results) == 1



"""Tests for IcebergRestCatalog integration with Snowflake macros."""

import pytest
from unittest.mock import Mock, MagicMock, patch
from types import SimpleNamespace

from dbt.adapters.snowflake.catalogs._iceberg_rest import (
    IcebergRestCatalogIntegration,
    IcebergRestCatalogRelation,
)
from dbt.adapters.catalogs import InvalidCatalogIntegrationConfigError


class TestIcebergRestCatalogIntegration:
    """Test integration between IcebergRestCatalog and Snowflake macros."""

    def setup_method(self):
        """Set up test fixtures."""
        self.mock_adapter = Mock()
        self.mock_target = Mock()
        self.mock_target.database = "original_database"

    def test_missing_catalog_linked_database_raises_error(self):
        """Error when adapter_properties lacks catalog_linked_database."""
        config = SimpleNamespace(
            name="test_iceberg_rest",
            catalog_name="POLARIS",
            catalog_type="iceberg_rest",
            external_volume="s3_volume",
            adapter_properties={},
        )

        with pytest.raises(InvalidCatalogIntegrationConfigError):
            IcebergRestCatalogIntegration(config)

    def test_catalog_relation_has_catalog_linked_database_attribute(self):
        """Test that catalog relation includes catalog_linked_database for macro usage."""
        config = SimpleNamespace(
            name="test_iceberg_rest",
            catalog_name="POLARIS",
            catalog_type="iceberg_rest",
            external_volume="s3_volume",
            adapter_properties={
                "catalog_linked_database": "custom_database",
            },
        )

        integration = IcebergRestCatalogIntegration(config)

        # Mock model config
        model = Mock()
        model.config = {}
        model.schema = "test_schema"
        model.identifier = "test_table"

        relation = integration.build_relation(model)

        # Verify the relation has the catalog_linked_database attribute
        assert hasattr(relation, "catalog_linked_database")
        assert relation.catalog_linked_database == "custom_database"

    def test_macro_integration_with_catalog_linked_database_set(self):
        """Test macro behavior when catalog_linked_database is set."""
        # Create a catalog relation with catalog_linked_database set
        relation = IcebergRestCatalogRelation(
            catalog_name="POLARIS",
            catalog_linked_database="custom_database",
            external_volume="test_volume",
        )

        # Test that hasattr returns True for catalog_linked_database
        assert hasattr(relation, "catalog_linked_database")
        assert relation.catalog_linked_database == "custom_database"

        # Simulate the macro logic from get_custom_name.sql
        # {%- if catalog_relation is not none and hasattr(catalog_relation, 'catalog_linked_database') -%}
        if relation is not None and hasattr(relation, "catalog_linked_database"):
            # This would return catalog_name in the actual macro
            result = relation.catalog_name
        else:
            # This would return target.database in the actual macro
            result = "target_database"

        assert result == "POLARIS"

    def test_macro_integration_without_catalog_relation(self):
        """Test macro behavior when catalog_relation is None."""
        relation = None

        # Simulate the macro logic from get_custom_name.sql
        # {%- if catalog_relation is not none and hasattr(catalog_relation, 'catalog_linked_database') -%}
        if relation is not None and hasattr(relation, "catalog_linked_database"):
            # This would return catalog_name in the actual macro
            result = relation.catalog_name
        else:
            # This would return target.database in the actual macro
            result = "target_database"

        assert result == "target_database"

    def test_integration_with_environment_variable_for_macro(self):
        """Test full integration with environment variable that would be used by macro."""
        import os

        config = SimpleNamespace(
            name="test_iceberg_rest",
            catalog_name="POLARIS",
            catalog_type="iceberg_rest",
            external_volume="s3_volume",
            adapter_properties={
                "catalog_linked_database": "catalog_linked_database",
            },
        )

        integration = IcebergRestCatalogIntegration(config)

        # Mock model config
        model = Mock()
        model.config = {}
        model.schema = "test_schema"
        model.identifier = "test_table"

        relation = integration.build_relation(model)

        # Verify the relation has the correct catalog_linked_database from env var
        assert relation.catalog_linked_database == "catalog_linked_database"

        # Simulate the macro check
        assert hasattr(relation, "catalog_linked_database")

    def test_catalog_relation_all_attributes_present(self):
        """Test that all expected attributes are present on the catalog relation."""
        relation = IcebergRestCatalogRelation(
            catalog_name="POLARIS",
            catalog_linked_database="custom_db",
            external_volume="test_volume",
            auto_refresh=True,
        )

        # Verify all attributes that might be used by macros
        assert hasattr(relation, "catalog_name")
        assert hasattr(relation, "catalog_linked_database")
        assert hasattr(relation, "external_volume")
        assert hasattr(relation, "auto_refresh")
        assert hasattr(relation, "catalog_type")
        assert hasattr(relation, "table_format")
        assert hasattr(relation, "file_format")

        # Verify values
        assert relation.catalog_name == "POLARIS"
        assert relation.catalog_linked_database == "custom_db"
        assert relation.external_volume == "test_volume"
        assert relation.auto_refresh is True



import os
import pytest
from unittest import mock

from dbt.adapters.snowflake.impl import SnowflakeAdapter
from dbt.adapters.capability import Capability, CapabilityDict
from dbt.cli.main import dbtRunner


freshness_via_metadata_schema_yml = """
sources:
  - name: test_source
    freshness:
      warn_after: {count: 10, period: hour}
      error_after: {count: 1, period: day}
    schema: "{{ env_var('DBT_GET_LAST_RELATION_TEST_SCHEMA') }}"
    tables:
      - name: test_table
"""

freshness_metadata_schema_batch_yml = """
sources:
  - name: test_source
    freshness:
      warn_after: {count: 10, period: hour}
      error_after: {count: 1, period: day}
    schema: "{{ env_var('DBT_GET_LAST_RELATION_TEST_SCHEMA') }}"
    tables:
      - name: test_table
      - name: test_table2
      - name: test_table_with_loaded_at_field
        loaded_at_field: my_loaded_at_field
"""


class SetupGetLastRelationModified:
    @pytest.fixture(scope="class", autouse=True)
    def set_env_vars(self, project):
        os.environ["DBT_GET_LAST_RELATION_TEST_SCHEMA"] = project.test_schema
        yield
        del os.environ["DBT_GET_LAST_RELATION_TEST_SCHEMA"]

    @pytest.fixture(scope="class")
    def custom_schema(self, project, set_env_vars):
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=project.database, schema=os.environ["DBT_GET_LAST_RELATION_TEST_SCHEMA"]
            )
            project.adapter.drop_schema(relation)
            project.adapter.create_schema(relation)

        yield relation.schema

        with project.adapter.connection_named("__test"):
            project.adapter.drop_schema(relation)


class TestGetLastRelationModified(SetupGetLastRelationModified):
    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": freshness_via_metadata_schema_yml}

    def test_get_last_relation_modified(self, project, set_env_vars, custom_schema):
        project.run_sql(
            f"create table {custom_schema}.test_table (id integer autoincrement, name varchar(100) not null);"
        )

        warning_or_error = False

        def probe(e):
            nonlocal warning_or_error
            if e.info.level in ["warning", "error"]:
                warning_or_error = True

        runner = dbtRunner(callbacks=[probe])
        runner.invoke(["source", "freshness"])

        # The 'source freshness' command should succeed without warnings or errors.
        assert not warning_or_error


class TestGetLastRelationModifiedBatch(SetupGetLastRelationModified):
    @pytest.fixture(scope="class")
    def models(self):
        return {"schema.yml": freshness_metadata_schema_batch_yml}

    def get_freshness_result_for_table(self, table_name, results):
        for result in results:
            if result.node.name == table_name:
                return result
        return None

    def test_get_last_relation_modified_batch(self, project, set_env_vars, custom_schema):
        project.run_sql(
            f"create table {custom_schema}.test_table (id integer autoincrement, name varchar(100) not null);"
        )
        project.run_sql(
            f"create table {custom_schema}.test_table2 (id integer autoincrement, name varchar(100) not null);"
        )
        project.run_sql(
            f"create table {custom_schema}.test_table_with_loaded_at_field as (select 1 as id, timestamp '2009-09-15 10:59:43' as my_loaded_at_field);"
        )

        runner = dbtRunner()
        freshness_results_batch = runner.invoke(["source", "freshness"]).result

        assert len(freshness_results_batch) == 3
        test_table_batch_result = self.get_freshness_result_for_table(
            "test_table", freshness_results_batch
        )
        test_table2_batch_result = self.get_freshness_result_for_table(
            "test_table2", freshness_results_batch
        )
        test_table_with_loaded_at_field_batch_result = self.get_freshness_result_for_table(
            "test_table_with_loaded_at_field", freshness_results_batch
        )

        # Remove TableLastModifiedMetadataBatch and run freshness on same input without batch strategy
        capabilities_no_batch = CapabilityDict(
            {
                capability: support
                for capability, support in SnowflakeAdapter.capabilities().items()
                if capability != Capability.TableLastModifiedMetadataBatch
            }
        )
        with mock.patch.object(
            SnowflakeAdapter, "capabilities", return_value=capabilities_no_batch
        ):
            freshness_results = runner.invoke(["source", "freshness"]).result

        assert len(freshness_results) == 3
        test_table_result = self.get_freshness_result_for_table("test_table", freshness_results)
        test_table2_result = self.get_freshness_result_for_table("test_table2", freshness_results)
        test_table_with_loaded_at_field_result = self.get_freshness_result_for_table(
            "test_table_with_loaded_at_field", freshness_results
        )

        # assert results between batch vs non-batch freshness strategy are equivalent
        assert test_table_result.status == test_table_batch_result.status
        assert test_table_result.max_loaded_at == test_table_batch_result.max_loaded_at

        assert test_table2_result.status == test_table2_batch_result.status
        assert test_table2_result.max_loaded_at == test_table2_batch_result.max_loaded_at

        assert (
            test_table_with_loaded_at_field_batch_result.status
            == test_table_with_loaded_at_field_result.status
        )
        assert (
            test_table_with_loaded_at_field_batch_result.max_loaded_at
            == test_table_with_loaded_at_field_result.max_loaded_at
        )



import pytest
import os
from dbt.tests.util import run_dbt, check_relations_equal_with_relations

_MODELS__VIEW_1_SQL = """
{#
    We are running against a database that must be quoted.
    These calls ensure that we trigger an error if we're failing to quote at parse-time
#}
{% do adapter.already_exists(this.schema, this.table) %}
{% do adapter.get_relation(this.database, this.schema, this.table) %}
select * from {{ ref('seed') }}
"""

_MODELS__VIEW_2_SQL = """
  {{ config(database=var('alternate_db')) }}
select * from {{ ref('seed') }}
"""

_MODELS__SUBFOLDER__VIEW_3_SQL = """
select * from {{ ref('seed') }}
"""

_MODELS__SUBFOLDER__VIEW_4_SQL = """
{{
    config(database=var('alternate_db'))
}}

select * from {{ ref('seed') }}
"""

_SEEDS__SEED_CSV = """id,name
1,a
2,b
3,c
4,d
5,e
"""

ALT_DATABASE = os.getenv("SNOWFLAKE_TEST_ALT_DATABASE")


class BaseOverrideDatabaseSnowflake:
    @pytest.fixture(scope="class")
    def seeds(self):
        return {"seed.csv": _SEEDS__SEED_CSV}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "view_1.sql": _MODELS__VIEW_1_SQL,
            "view_2.sql": _MODELS__VIEW_2_SQL,
            "subfolder": {
                "view_3.sql": _MODELS__SUBFOLDER__VIEW_3_SQL,
                "view_4.sql": _MODELS__SUBFOLDER__VIEW_4_SQL,
            },
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "vars": {
                "alternate_db": ALT_DATABASE,
            },
            "quoting": {
                "database": True,
            },
        }

    @pytest.fixture(scope="function")
    def clean_up(self, project):
        yield
        with project.adapter.connection_named("__test"):
            relation = project.adapter.Relation.create(
                database=ALT_DATABASE, schema=project.test_schema
            )
            project.adapter.drop_schema(relation)

    @staticmethod
    def cap_name(name):
        return name.upper()


class TestModelOverrideSnowflake(BaseOverrideDatabaseSnowflake):
    def run_database_override(self, project):
        run_dbt(["seed"])
        result = run_dbt(["run"])
        assert len(result) == 4
        check_relations_equal_with_relations(
            project.adapter,
            [
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("seed")
                ),
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("view_1")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_2"),
                ),
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("view_3")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_4"),
                ),
            ],
        )

    def test_snowflake_database_override(self, project, clean_up):
        self.run_database_override(project)


class TestProjectSeedOverrideSnowflake(BaseOverrideDatabaseSnowflake):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "vars": {
                "alternate_db": ALT_DATABASE,
            },
            "seeds": {"database": ALT_DATABASE},
        }

    def run_database_override(self, project):
        run_dbt(["seed"])
        assert len(run_dbt(["run"])) == 4
        check_relations_equal_with_relations(
            project.adapter,
            [
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("seed"),
                ),
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("view_1")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_2"),
                ),
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("view_3")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_4"),
                ),
            ],
        )

    def test_snowflake_database_override(self, project, clean_up):
        self.run_database_override(project)


class BaseProjectModelOverrideSnowflake(BaseOverrideDatabaseSnowflake):
    def run_database_override(self, project):
        run_dbt(["seed"])
        result = run_dbt(["run"])
        assert len(result) == 4
        check_relations_equal_with_relations(
            project.adapter,
            [
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("seed")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_1"),
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_2"),
                ),
                project.adapter.Relation.create(
                    schema=project.test_schema, identifier=self.cap_name("view_3")
                ),
                project.adapter.Relation.create(
                    database=ALT_DATABASE,
                    schema=project.test_schema,
                    identifier=self.cap_name("view_4"),
                ),
            ],
        )


class TestProjectModelOverrideSnowflake(BaseProjectModelOverrideSnowflake):
    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "models": {
                "database": ALT_DATABASE,
                "test": {"subfolder": {"database": "{{ target.database }}"}},
            },
            "vars": {
                "alternate_db": ALT_DATABASE,
            },
        }

    def test_snowflake_database_override(self, project, clean_up):
        self.run_database_override(project)



seed_csv = """
id,first_name,last_name,email,gender,ip_address
1,Jack,Hunter,jhunter0@pbs.org,Male,59.80.20.168
2,Kathryn,Walker,kwalker1@ezinearticles.com,Female,194.121.179.35
3,Gerald,Ryan,gryan2@com.com,Male,11.3.212.243
""".lstrip()

seed_agg_csv = """
last_name,count
Hunter,2
Walker,2
Ryan,2
""".lstrip()



import pytest

from dbt.tests.util import run_dbt, run_dbt_and_capture

from tests.functional.relation_tests.dynamic_table_tests import models
from tests.functional.utils import query_relation_type


class TestDynamicTableFullRefreshCustomSchema:

    @pytest.fixture(scope="class", autouse=True)
    def models(self):
        return {
            "simple_model.sql": models.SIMPLE_MODEL,
            "custom_schema_dynamic_table.sql": models.DYNAMIC_TABLE_CUSTOM_SCHEMA,
        }

    def test_dynamic_table_full_refresh_with_custom_schema(self, project):
        """Test that dynamic table full refresh works with custom schema configuration."""
        # Initial run to create the dynamic table
        run_dbt(["seed"])
        run_dbt(["run"])

        # Run full refresh - this was failing before the fix
        run_dbt(["run", "--full-refresh"])
        run_dbt(["run", "--full-refresh"])



"""Unit tests for platform_detection_timeout_seconds configuration"""

import pytest
from dbt.adapters.snowflake.connections import SnowflakeCredentials


class TestPlatformDetectionTimeout:
    """Test suite for platform_detection_timeout_seconds configuration"""

    def test_default_platform_detection_timeout(self):
        """Test that platform_detection_timeout_seconds defaults to 0.0"""
        creds = SnowflakeCredentials(
            account="test_account",
            user="test_user",
            password="test_password",
            database="test_db",
            warehouse="test_warehouse",
            schema="test_schema",
        )
        assert creds.platform_detection_timeout_seconds == 0.0

    def test_custom_platform_detection_timeout(self):
        """Test that platform_detection_timeout_seconds accepts custom values"""
        creds = SnowflakeCredentials(
            account="test_account",
            user="test_user",
            password="test_password",
            database="test_db",
            warehouse="test_warehouse",
            schema="test_schema",
            platform_detection_timeout_seconds=5.5,
        )
        assert creds.platform_detection_timeout_seconds == 5.5

    def test_platform_detection_timeout_in_connection_keys(self):
        """Test that platform_detection_timeout_seconds is included in connection keys"""
        creds = SnowflakeCredentials(
            account="test_account",
            user="test_user",
            password="test_password",
            database="test_db",
            warehouse="test_warehouse",
            schema="test_schema",
        )
        assert "platform_detection_timeout_seconds" in creds._connection_keys()

    @pytest.mark.parametrize("timeout_value", [0.0, 1.0, 5.5, 10.0, 30.0])
    def test_various_timeout_values(self, timeout_value):
        """Test that platform_detection_timeout_seconds accepts various float values"""
        creds = SnowflakeCredentials(
            account="test_account",
            user="test_user",
            password="test_password",
            database="test_db",
            warehouse="test_warehouse",
            schema="test_schema",
            platform_detection_timeout_seconds=timeout_value,
        )
        assert creds.platform_detection_timeout_seconds == timeout_value



from unittest import mock

import dbt.adapters.snowflake.__version__

from dbt.adapters.snowflake.impl import SnowflakeAdapter
from dbt.adapters.base.relation import AdapterTrackingRelationInfo


def test_telemetry_with_snowflake_details():
    mock_model_config = mock.MagicMock()
    mock_model_config._extra = mock.MagicMock()
    mock_model_config._extra = {
        "adapter_type": "snowflake",
        "table_format": "iceberg",
    }

    res = SnowflakeAdapter.get_adapter_run_info(mock_model_config)

    assert res.adapter_name == "snowflake"
    assert res.base_adapter_version == dbt.adapters.__about__.version
    assert res.adapter_version == dbt.adapters.snowflake.__version__.version
    assert res.model_adapter_details == {
        "adapter_type": "snowflake",
        "table_format": "iceberg",
    }

    assert type(res) is AdapterTrackingRelationInfo



SCHEMA__CONTROL = """
version: 2
models:
  - name: colors
    columns:
      - name: color
        data_tests:
          - not_null
"""


SCHEMA__EXPLICIT_WAREHOUSE = """
version: 2
models:
  - name: colors
    columns:
      - name: color
        data_tests:
          - not_null:
              config:
                snowflake_warehouse: DBT_TESTING_ALT
"""


SCHEMA__NOT_NULL = """
version: 2
models:
  - name: facts
    columns:
      - name: value
        data_tests:
          - not_null:
              config:
                snowflake_warehouse: DBT_TESTING_ALT
"""


SCHEMA__RELATIONSHIPS = """
version: 2
models:
  - name: facts
    columns:
      - name: color
        data_tests:
          - relationships:
              to: ref('my_colors')
              field: color
              config:
                snowflake_warehouse: DBT_TESTING_ALT
"""


SCHEMA__ACCEPTED_VALUES = """
version: 2
models:
  - name: colors
    columns:
      - name: color
        data_tests:
          - accepted_values:
              values: ['blue', 'red', 'green']
              config:
                snowflake_warehouse: DBT_TESTING_ALT
"""


SEED__COLORS = """
color
blue
green
red
yellow
""".strip()


# record 10 is missing a value
# record 7 has a color that's not on COLORS
SEED__FACTS = """
id,color,value
1,blue,10
2,red,20
3,green,30
4,yellow,40
5,blue,50
6,red,60
7,orange,70
8,green,80
9,yellow,90
10,blue,
""".strip()



from tests.functional.relation_tests.base import RelationOperation


class TestTable(RelationOperation):

    def test_get_create_backup_and_rename_intermediate_sql(self, project):
        args = {
            "database": project.database,
            "schema": project.test_schema,
            "identifier": "my_table",
            "relation_type": "table",
        }
        expected_statement = (
            f"alter table {project.database}.{project.test_schema}.my_table "
            f"rename to {project.database}.{project.test_schema}.my_table__dbt_backup"
        )
        self.assert_operation(project, "test__get_create_backup_sql", args, expected_statement)

        expected_statement = (
            f"alter table {project.database}.{project.test_schema}.my_table__dbt_tmp "
            f"rename to {project.database}.{project.test_schema}.my_table"
        )
        self.assert_operation(
            project, "test__get_rename_intermediate_sql", args, expected_statement
        )



"""Unit tests for SnowflakeColumn collation support."""

import importlib.util
from pathlib import Path
import sys
import unittest

from dbt.adapters.snowflake.column import SnowflakeColumn


class TestSnowflakeColumnCollation(unittest.TestCase):
    """Test cases for collation parsing and generation in SnowflakeColumn."""

    def test_parse_varchar_with_collation(self):
        """Test parsing VARCHAR with COLLATE clause."""
        column = SnowflakeColumn.from_description(
            "test_col", "VARCHAR(16777216) COLLATE 'en-ci-rtrim'"
        )

        assert column.name == "test_col"
        assert column.dtype == "VARCHAR"
        assert column.char_size == 16777216
        assert column.collation == "en-ci-rtrim"

    def test_parse_varchar_without_collation(self):
        """Test parsing VARCHAR without COLLATE clause."""
        column = SnowflakeColumn.from_description("test_col", "VARCHAR(100)")

        assert column.name == "test_col"
        assert column.dtype == "VARCHAR"
        assert column.char_size == 100
        assert column.collation is None

    def test_parse_text_with_collation(self):
        """Test parsing TEXT with COLLATE clause."""
        column = SnowflakeColumn.from_description("test_col", "TEXT COLLATE 'es-ai'")

        assert column.name == "test_col"
        assert column.dtype == "TEXT"
        assert column.collation == "es-ai"
        assert column.expanded_data_type == "character varying(16777216) collate 'es-ai'"

    def test_parse_character_varying_with_collation(self):
        """Test parsing CHARACTER VARYING with COLLATE clause."""
        column = SnowflakeColumn.from_description(
            "test_col", "CHARACTER VARYING(1000) COLLATE 'fr'"
        )

        assert column.name == "test_col"
        assert column.dtype == "CHARACTER VARYING"
        assert column.char_size == 1000
        assert column.collation == "fr"

    def test_parse_number_without_collation(self):
        """Test parsing NUMBER (non-string type) has no collation."""
        column = SnowflakeColumn.from_description("test_col", "NUMBER(38,0)")

        assert column.name == "test_col"
        assert column.dtype == "NUMBER"
        assert column.numeric_precision == 38
        assert column.numeric_scale == 0
        assert column.collation is None

    def test_data_type_with_collation(self):
        """Test that expanded_data_type property includes collation for strings."""
        column = SnowflakeColumn.from_description(
            "test_col", "VARCHAR(16777216) COLLATE 'en-ci-rtrim'"
        )

        # data_type should not include collation; expanded_data_type should
        assert column.data_type == "character varying(16777216)"
        assert column.expanded_data_type == "character varying(16777216) collate 'en-ci-rtrim'"

    def test_data_type_without_collation(self):
        """Test that expanded_data_type property works without collation."""
        column = SnowflakeColumn.from_description("test_col", "VARCHAR(100)")

        # data_type should not include collation
        expected = "character varying(100)"
        assert column.data_type == expected
        assert column.expanded_data_type == expected

    def test_data_type_number_ignores_collation(self):
        """Test that non-string types don't include collation even if set."""
        column = SnowflakeColumn.from_description("test_col", "NUMBER(10,2)")
        # Manually set collation (shouldn't happen in practice)
        column.collation = "en-ci"

        # data_type/expanded_data_type should not include collation for non-string types
        expected = "NUMBER(10,2)"
        assert column.data_type == expected
        assert column.expanded_data_type == expected

    def test_parse_collation_case_insensitive(self):
        """Test that COLLATE keyword is case-insensitive."""
        # Test uppercase
        column1 = SnowflakeColumn.from_description("test_col", "VARCHAR(100) COLLATE 'en-ci'")
        assert column1.collation == "en-ci"

        # Test lowercase
        column2 = SnowflakeColumn.from_description("test_col", "VARCHAR(100) collate 'en-ci'")
        assert column2.collation == "en-ci"

    def test_parse_varchar_max_size_with_collation(self):
        """Test parsing VARCHAR with maximum size (134217728) and collation."""
        column = SnowflakeColumn.from_description(
            "test_col", "VARCHAR(134217728) COLLATE 'en-ci-rtrim'"
        )

        assert column.name == "test_col"
        assert column.dtype == "VARCHAR"
        assert column.char_size == 134217728
        assert column.collation == "en-ci-rtrim"

    def test_data_type_preserves_collation_after_size_change(self):
        """Test that collation is preserved when generating expanded_data_type."""
        # This simulates what happens during schema change detection
        column = SnowflakeColumn.from_description(
            "test_col", "VARCHAR(16777216) COLLATE 'en-ci-rtrim'"
        )

        # When dbt detects a size change, it should preserve collation
        assert "collate 'en-ci-rtrim'" in column.expanded_data_type

    def test_parse_text_without_size_with_collation(self):
        """Test parsing TEXT (no size) with collation."""
        column = SnowflakeColumn.from_description("test_col", "TEXT COLLATE 'utf8'")

        assert column.name == "test_col"
        assert column.dtype == "TEXT"
        assert column.char_size is None
        assert column.collation == "utf8"
        assert "collate 'utf8'" in column.expanded_data_type

    def test_parse_array_type_no_collation(self):
        """Test that ARRAY types don't have collation parsing."""
        column = SnowflakeColumn.from_description("test_col", "ARRAY")

        assert column.name == "test_col"
        assert column.dtype == "ARRAY"
        assert column.collation is None

    def test_parse_object_type_no_collation(self):
        """Test that OBJECT types don't have collation parsing."""
        column = SnowflakeColumn.from_description("test_col", "OBJECT")

        assert column.name == "test_col"
        assert column.dtype == "OBJECT"
        assert column.collation is None



import pytest

from pathlib import Path

from dbt.tests.util import run_dbt, run_dbt_and_capture, write_file


_SEEDS_SIMPLE_SEED = """
id,value,event_date
1,foo,2025-01-01
2,bar,2025-01-01
""".lstrip()

_SEEDS_SIMPLE_SEED_UPDATE = """
id,value,event_date
1,foo,2025-01-01
2,cat,2025-03-01
""".lstrip()

_MODELS_INCREMENTAL_MODEL = """
{{ config(
    materialized='incremental',
    unique_key='id',
    incremental_strategy='insert_overwrite',
) }}

select
    id, value, event_date
from {{ ref('seed') }}
    {% if is_incremental() %}
        where event_date > '2025-02-01'
    {% endif %}
"""

_MODELS_INCREMENTAL_MODEL_WITH_COLUMNS = """
{{ config(
    materialized='incremental',
    unique_key='id',
    incremental_strategy='insert_overwrite',
    overwrite_columns=['id', 'value']
) }}

select
    id, value, event_date
from {{ ref('seed') }}
    {% if is_incremental() %}
        where event_date > '2025-02-01'
    {% endif %}
"""


class TestInsertOverwriteIncremental:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "incremental.sql": _MODELS_INCREMENTAL_MODEL,
            "incremental_with_cols.sql": _MODELS_INCREMENTAL_MODEL_WITH_COLUMNS,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {
            "seed.csv": _SEEDS_SIMPLE_SEED,
        }

    def test_insert_overwrite_incremental(self, project):
        results = run_dbt(["seed"])
        assert len(results) == 1

        results = run_dbt(["run"])
        assert len(results) == 2

        seed_file = project.project_root / Path("seeds") / Path("seed.csv")
        write_file(_SEEDS_SIMPLE_SEED_UPDATE, seed_file)

        results = run_dbt(["seed"])
        assert len(results) == 1

        results = run_dbt(["run"])
        assert len(results) == 2

        run_results, output = run_dbt_and_capture(
            ["show", "--inline", "select * from {{ ref('incremental') }}"]
        )
        assert run_results[0].adapter_response["rows_affected"] == 1
        assert "cat" in output
        assert "2025-03-01" in output
        assert "foo" not in output
        assert "2025-01-01" not in output
        assert run_results[0].adapter_response["code"] == "SUCCESS"

        run_results, output = run_dbt_and_capture(
            [
                "show",
                "--inline",
                "select * from {{ ref('incremental_with_cols') }} where value = 'cat'",
            ]
        )
        assert run_results[0].adapter_response["rows_affected"] == 1
        assert "cat" in output
        assert "2025-03-01" not in output and "2025-01-01" not in output
        assert run_results[0].adapter_response["code"] == "SUCCESS"



import pytest
from dbt.tests.adapter.simple_seed.test_seed_type_override import BaseSimpleSeedColumnOverride
from dbt.tests.adapter.utils.base_utils import run_dbt

_SCHEMA_YML = """
version: 2
seeds:
- name: seed_enabled
  columns:
  - name: birthday
    data_tests:
    - column_type:
        type: character varying(16777216)
  - name: seed_id
    data_tests:
    - column_type:
        type: FLOAT

- name: seed_tricky
  columns:
  - name: seed_id
    data_tests:
    - column_type:
        type: NUMBER(38,0)
  - name: seed_id_str
    data_tests:
    - column_type:
        type: character varying(16777216)
  - name: a_bool
    data_tests:
    - column_type:
        type: BOOLEAN
  - name: looks_like_a_bool
    data_tests:
    - column_type:
        type: character varying(16777216)
  - name: a_date
    data_tests:
    - column_type:
        type: TIMESTAMP_NTZ
  - name: looks_like_a_date
    data_tests:
    - column_type:
        type: character varying(16777216)
  - name: relative
    data_tests:
    - column_type:
        type: character varying(16777216)
  - name: weekday
    data_tests:
    - column_type:
        type: character varying(16777216)
""".lstrip()


class TestSimpleSeedColumnOverride(BaseSimpleSeedColumnOverride):
    @pytest.fixture(scope="class")
    def schema(self):
        return "simple_seed"

    @pytest.fixture(scope="class")
    def models(self):
        return {"models-snowflake.yml": _SCHEMA_YML}

    @staticmethod
    def seed_enabled_types():
        return {
            "seed_id": "FLOAT",
            "birthday": "TEXT",
        }

    @staticmethod
    def seed_tricky_types():
        return {
            "seed_id_str": "TEXT",
            "looks_like_a_bool": "TEXT",
            "looks_like_a_date": "TEXT",
        }

    def test_snowflake_simple_seed_with_column_override_snowflake(self, project):
        seed_results = run_dbt(["seed"])
        assert len(seed_results) == 2
        test_results = run_dbt(["test"])
        assert len(test_results) == 10



"""
Please follow the instructions in test_oauth.py for instructions on how to set up
the security integration required to retrieve a JWT from Snowflake.
"""

import pytest
import os
from dbt.tests.util import run_dbt, check_relations_equal

from dbt.adapters.snowflake import SnowflakeCredentials

_MODELS__MODEL_1_SQL = """
select 1 as id
"""


_MODELS__MODEL_2_SQL = """
select 2 as id
"""


_MODELS__MODEL_3_SQL = """
select * from {{ ref('model_1') }}
union all
select * from {{ ref('model_2') }}
"""


_MODELS__MODEL_4_SQL = """
select 1 as id
union all
select 2 as id
"""


class TestSnowflakeJWT:
    """Tests that setting authenticator: jwt allows setting token to a plain JWT
    that will be passed into the Snowflake connection without modification."""

    @pytest.fixture(scope="class", autouse=True)
    def access_token(self):
        """Because JWTs are short-lived, we need to get a fresh JWT via the refresh
        token flow before running the test.

        This fixture leverages the existing SnowflakeCredentials._get_access_token
        method to retrieve a valid JWT from Snowflake.
        """
        client_id = os.getenv("SNOWFLAKE_TEST_OAUTH_CLIENT_ID")
        client_secret = os.getenv("SNOWFLAKE_TEST_OAUTH_CLIENT_SECRET")
        refresh_token = os.getenv("SNOWFLAKE_TEST_OAUTH_REFRESH_TOKEN")

        credentials = SnowflakeCredentials(
            account=os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
            database="",
            schema="",
            authenticator="oauth",
            oauth_client_id=client_id,
            oauth_client_secret=client_secret,
            token=refresh_token,
        )

        yield credentials._get_access_token()

    @pytest.fixture(scope="class", autouse=True)
    def dbt_profile_target(self, access_token):
        """A dbt_profile that has authenticator set to JWT, and token set to
        a JWT accepted by Snowflake. Also omits the user, as the user attribute
        is optional when the authenticator is set to JWT.
        """
        return {
            "type": "snowflake",
            "threads": 4,
            "account": os.getenv("SNOWFLAKE_TEST_ACCOUNT"),
            "database": os.getenv("SNOWFLAKE_TEST_DATABASE"),
            "warehouse": os.getenv("SNOWFLAKE_TEST_WAREHOUSE"),
            "authenticator": "jwt",
            "token": access_token,
        }

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "model_1.sql": _MODELS__MODEL_1_SQL,
            "model_2.sql": _MODELS__MODEL_2_SQL,
            "model_3.sql": _MODELS__MODEL_3_SQL,
            "model_4.sql": _MODELS__MODEL_4_SQL,
        }

    def test_snowflake_basic(self, project):
        run_dbt()
        check_relations_equal(project.adapter, ["MODEL_3", "MODEL_4"])



import pytest
from dbt.tests.util import run_dbt, check_relations_equal

_MODELS__DEPENDENT_MODEL_SQL = """
{% if var('dependent_type', 'view') == 'view' %}
    {{ config(materialized='view') }}
{% else %}
    {{ config(materialized='table') }}
{% endif %}

select * from {{ ref('base_table') }}
"""


_MODELS__BASE_TABLE_SQL = """
{{ config(materialized='table') }}
select *
    {% if var('add_table_field', False) %}
        , 1 as new_field
    {% endif %}

from {{ ref('people') }}
"""

_SEEDS__PEOPLE_CSV = """id,name
1,Drew
2,Jake
3,Connor
"""


class TestSnowflakeLateBindingViewDependency:
    @pytest.fixture(scope="class")
    def models(self):
        return {
            "dependent_model.sql": _MODELS__DEPENDENT_MODEL_SQL,
            "base_table.sql": _MODELS__BASE_TABLE_SQL,
        }

    @pytest.fixture(scope="class")
    def seeds(self):
        return {"people.csv": _SEEDS__PEOPLE_CSV}

    @pytest.fixture(scope="class")
    def project_config_update(self):
        return {
            "seeds": {
                "quote_columns": False,
            },
            "quoting": {"schema": False, "identifier": False},
        }

    @pytest.fixture(scope="class", autouse=True)
    def setup_method(self, project):
        results = run_dbt(["seed"])
        assert len(results) == 1
        results = run_dbt(["run"])
        assert len(results) == 2
        check_relations_equal(project.adapter, ["PEOPLE", "BASE_TABLE"])
        check_relations_equal(project.adapter, ["PEOPLE", "DEPENDENT_MODEL"])

    def check_result(self, project, results, expected_types):
        for result in results:
            node_name = result.node.name
            node_type = result.node.config.materialized
            assert node_type == expected_types[node_name]

    """
    Snowflake views are not bound to the relations they select from. A Snowflake view
    can have entirely invalid SQL if, for example, the table it selects from is dropped
    and recreated with a different schema. In these scenarios, Snowflake will raise an error if:
    1) The view is queried
    2) The view is altered

    dbt's logic should avoid running these types of queries against views in situations
    where they _may_ have invalid definitions. These tests assert that views are handled
    correctly in various different scenarios
    """

    def test__snowflake__changed_table_schema_for_downstream_view(self, project):
        run_dbt(["seed"])
        # Change the schema of base_table, assert that dependent_model doesn't fail
        results = run_dbt(["run", "--vars", "{add_table_field: true, dependent_type: view}"])
        assert len(results) == 2
        check_relations_equal(project.adapter, ["BASE_TABLE", "DEPENDENT_MODEL"])

    """
    This test is similar to the one above, except the downstream model starts as a view, and
    then is changed to be a table. This checks that the table materialization does not
    errantly rename a view that might have an invalid definition, which would cause an error
    """

    def test__snowflake__changed_table_schema_for_downstream_view_changed_to_table(self, project):
        run_dbt(["seed"])
        results = run_dbt(["run"])
        expected_types = {"base_table": "table", "dependent_model": "view"}
        # ensure that the model actually was materialized as a view
        self.check_result(project, results, expected_types)
        results = run_dbt(["run", "--vars", "{add_table_field: true, dependent_type: table}"])
        assert len(results) == 2
        check_relations_equal(project.adapter, ["BASE_TABLE", "DEPENDENT_MODEL"])
        expected_types = {"base_table": "table", "dependent_model": "table"}
        # ensure that the model actually was materialized as a table
        self.check_result(project, results, expected_types)



import os
import pytest
from dbt.tests.adapter.catalog_integrations.test_catalog_integration import (
    BaseCatalogIntegrationValidation,
)
from dbt.tests.util import run_dbt

MODEL__BASIC_ICEBERG_TABLE = """
                            {{ config(materialized='table',
                             catalog_name='basic_iceberg_rest_catalog') }}
                            select 1 as id
                            """

MODEL__ICEBERG_TABLE_WITH_CATALOG_CONFIG = """
                            {{ config(materialized='table', catalog_name='basic_iceberg_rest_catalog',
                            target_file_size='16MB', max_data_extension_time_in_days=1, auto_refresh='true') }}
                            select 1 as id
                            """

MODEL__INCREMENTAL_ICEBERG_REST = """
{{
  config(
    materialized='incremental',
    catalog_name='basic_iceberg_rest_catalog',
    incremental_strategy='merge',
    unique_key="id",
  )
}}
select * from {{ ref('basic_iceberg_table') }}

{% if is_incremental() %}
where id > 2
{% endif %}
"""

MODEL__INCREMENTAL_ICEBERG_REST_INSERT_OVERWRITE = """
{{
  config(
    materialized='incremental',
    catalog_name='basic_iceberg_rest_catalog',
    incremental_strategy='insert_overwrite',
    unique_key="id",
  )
}}
select * from {{ ref('basic_iceberg_table') }}

{% if is_incremental() %}
where id > 2
{% endif %}
"""


class TestSnowflakeIcebergRestCatalogIntegration(BaseCatalogIntegrationValidation):

    @pytest.fixture(scope="class")
    def catalogs(self):
        return {
            "catalogs": [
                {
                    "name": "basic_iceberg_rest_catalog",
                    "active_write_integration": "iceberg_rest_catalog_with_linked_db_integration",
                    "write_integrations": [
                        {
                            "name": "iceberg_rest_catalog_with_linked_db_integration",
                            "catalog_type": "iceberg_rest",
                            "table_format": "iceberg",
                            "adapter_properties": {
                                "catalog_linked_database": os.getenv(
                                    "SNOWFLAKE_TEST_CATALOG_LINKED_DATABASE"
                                ),
                                # No catalog_linked_database_type means standard CTAS is used
                                "max_data_extension_time_in_days": 1,
                                "target_file_size": "AUTO",
                                "auto_refresh": "true",
                            },
                        }
                    ],
                },
            ]
        }

    # can only use alphanumeric characters in schema names in catalog linked databases
    # until it's GA
    @pytest.fixture(scope="class")
    def unique_schema(self, request, prefix) -> str:
        test_file = request.module.__name__
        # We only want the last part of the name
        test_file = test_file.split(".")[-1]
        unique_schema = f"{prefix}_{test_file}"
        return unique_schema.replace("_", "")

    @pytest.fixture(scope="class")
    def models(self):
        return {
            "models": {
                "basic_iceberg_table.sql": MODEL__BASIC_ICEBERG_TABLE,
                "iceberg_table_with_catalog_config.sql": MODEL__ICEBERG_TABLE_WITH_CATALOG_CONFIG,
                "incremental_iceberg_rest.sql": MODEL__INCREMENTAL_ICEBERG_REST,
                "incremental_iceberg_rest_insert_overwrite.sql": MODEL__INCREMENTAL_ICEBERG_REST_INSERT_OVERWRITE,
            }
        }

    def test_basic_iceberg_rest_catalog_integration(self, project):
        result = run_dbt(["run"])
        assert len(result) == 4
        run_dbt(["run"])


class TestSnowflakeIcebergRestGlueCatalogIntegration(BaseCatalogIntegrationValidation):

    @pytest.fixture(scope="class")
    def catalogs(self):
        return {
            "catalogs": [
                {
                    "name": "glue_iceberg_rest_catalog",
                    "active_write_integration": "glue_iceberg_rest_catalog_integration",
                    "write_integrations": [
                        {
                            "name": "glue_iceberg_rest_catalog_integration",
                            "catalog_type": "iceberg_rest",
                            "table_format": "iceberg",
                            "adapter_properties": {
                                "catalog_linked_database": os.getenv(
                                    "SNOWFLAKE_TEST_CATALOG_LINKED_DATABASE_GLUE"
                                ),
                                "catalog_linked_database_type": "glue",  # Glue requires 4-step process
                                "max_data_extension_time_in_days": 1,
                                "target_file_size": "AUTO",
                                "auto_refresh": "true",
                            },
                        }
                    ],
                },
            ]
        }

    @pytest.fixture(scope="class")
    def project_config_update(self):
        # Force quoting for Glue CLD compatibility
        return {
            "quoting": {
                "database": False,
                "schema": True,
                "identifier": True,
            }
        }

    # @pytest.fixture(scope="class", autouse=True)
    # def setup_glue_schema(self, project):
    #     """Pre-create schema with quoted lowercase identifier for Glue CLD"""
    #     adapter = project.adapter
    #     glue_database = os.getenv("SNOWFLAKE_TEST_CATALOG_LINKED_DATABASE_GLUE")
    #     schema_name = project.test_schema.lower()

    #     # Create schema with quoted identifier to preserve lowercase
    #     create_schema_sql = f'CREATE SCHEMA IF NOT EXISTS {glue_database}."{schema_name}"'
    #     adapter.execute(create_schema_sql, fetch=False)

    #     yield

    #     # Cleanup: drop schema after test
    #     drop_schema_sql = f'DROP SCHEMA IF EXISTS {glue_database}."{schema_name}"'
    #     try:
    #         adapter.execute(drop_schema_sql, fetch=False)
    #     except:
    #         pass  # Ignore cleanup errors

    # AWS Glue requires lowercase identifiers and alphanumeric characters only
    @pytest.fixture(scope="class")
    def unique_schema(self, request, prefix) -> str:
        test_file = request.module.__name__
        # We only want the last part of the name
        test_file = test_file.split(".")[-1]
        unique_schema = f"{prefix}_{test_file}_glue"
        # Remove underscores and convert to lowercase for Glue compatibility
        return unique_schema.replace("_", "").lower()

    @pytest.fixture(scope="class")
    def models(self):
        # Use different catalog name for Glue models
        return {
            "models": {
                "glue_basic_iceberg_table.sql": """
                    {{ config(materialized='table', schema=target.schema.upper(),
                     catalog_name='glue_iceberg_rest_catalog') }}
                    select 1 as id, 'test' as name, 1.0 as price, '2021-01-01' as test_date
                """,
                "glue_iceberg_table_with_catalog_config.sql": """
                    {{ config(materialized='table', catalog_name='glue_iceberg_rest_catalog',
                    target_file_size='16MB', max_data_extension_time_in_days=1, auto_refresh='true') }}
                    select 1 as id
                """,
            }
        }

    def test_glue_iceberg_rest_catalog_integration(self, project):
        """Test Glue CLD with 4-step table creation process"""
        result = run_dbt(["run"])
        assert len(result) == 2
        # Run again to test update path
        run_dbt(["run"])



from dbt.tests.adapter.sample_mode.test_sample_mode import (
    BaseSampleModeTest,
)
import pytest


_input_model_sql = """
{{ config(materialized='table', event_time='event_time') }}
select 1 as id, to_timestamp_tz('2025-01-01 01:25:00-0') as event_time
union all
select 2 as id, to_timestamp_tz('2025-01-02 13:47:00-0') as event_time
union all
select 3 as id, to_timestamp_tz('2025-01-03 01:32:00-0') as event_time
"""


@pytest.mark.skip(reason="Sample mode is unsupported on 1.9")
class TestSnowflakeSampleMode(BaseSampleModeTest):
    @pytest.fixture(scope="class")
    def input_model_sql(self) -> str:
        return _input_model_sql



from dbt.tests.adapter.simple_snapshot.test_snapshot import BaseSnapshotCheck, BaseSimpleSnapshot
from dbt.tests.adapter.simple_snapshot.new_record_timestamp_mode import (
    BaseSnapshotNewRecordTimestampMode,
)
from dbt.tests.adapter.simple_snapshot.new_record_check_mode import BaseSnapshotNewRecordCheckMode


class TestSnapshot(BaseSimpleSnapshot):
    pass


class TestSnapshotCheck(BaseSnapshotCheck):
    pass


class TestSnapshotNewRecordTimestampMode(BaseSnapshotNewRecordTimestampMode):
    pass


class TestSnapshotNewRecordCheckMode(BaseSnapshotNewRecordCheckMode):
    pass



import pytest

from dbt.tests.util import run_dbt

from tests.functional.adapter.python_model_tests import _files


class TestTableType:
    @pytest.fixture(scope="class")
    def macros(self):
        return {"snowflake__test__describe_tables.sql": _files.MACRO__DESCRIBE_TABLES}

    @pytest.fixture(scope="class")
    def models(self):
        return {
            # <temporary>_<transient>_table
            "TRANSIENT_TRUE_TABLE.py": _files.TRANSIENT_TRUE_TABLE,
            "TRANSIENT_FALSE_TABLE.py": _files.TRANSIENT_FALSE_TABLE,
            "TRANSIENT_NONE_TABLE.py": _files.TRANSIENT_NONE_TABLE,
            "TRANSIENT_UNSET_TABLE.py": _files.TRANSIENT_UNSET_TABLE,
        }

    def test_expected_table_types_are_created(self, project):
        run_dbt(["run"])
        expected_table_types = {
            # (name, kind) - TABLE == permanent
            ("TRANSIENT_TRUE_TABLE", "TRANSIENT"),
            ("TRANSIENT_FALSE_TABLE", "TABLE"),
            ("TRANSIENT_NONE_TABLE", "TABLE"),
            ("TRANSIENT_UNSET_TABLE", "TRANSIENT"),
        }
        with project.adapter.connection_named("__test"):
            agate_table = project.adapter.execute_macro("snowflake__test__describe_tables")
        actual_table_types = {(row.get("name"), row.get("kind")) for row in agate_table.rows}
        assert actual_table_types == expected_table_types
